"""
Robert "Uncle Bob" Martin - Agile software metrics
https://en.wikipedia.org/wiki/Software_package_metrics

Efferent Coupling (Ce): Number of contracts that the contract depends on
Afferent Coupling (Ca): Number of contracts that depend on a contract
Instability (I): Ratio of efferent coupling to total coupling (Ce / (Ce + Ca))
Abstractness (A): Number of abstract contracts / total number of contracts
Distance from the Main Sequence (D):  abs(A + I - 1)

"""

from dataclasses import dataclass, field
from collections import OrderedDict
from slither.slithir.operations.high_level_call import HighLevelCall
from slither.core.declarations import Contract
from slither.utils.myprettytable import make_pretty_table, MyPrettyTable


@dataclass(slots=True)
class MartinContractMetrics:
    contract: Contract
    ca: int
    ce: int
    abstractness: float
    i: float = 0.0
    d: float = 0.0

    def __post_init__(self) -> None:
        if self.ce + self.ca > 0:
            self.i = float(self.ce / (self.ce + self.ca))
            self.d = float(abs(self.i - self.abstractness))

    def to_dict(self) -> dict:
        return {
            "Dependents": self.ca,
            "Dependencies": self.ce,
            "Instability": f"{self.i:.2f}",
            "Distance from main sequence": f"{self.d:.2f}",
        }


@dataclass(slots=True)
class SectionInfo:
    """Class to hold the information for a section of the report."""

    title: str
    pretty_table: MyPrettyTable
    txt: str


@dataclass(slots=True)
class MartinMetrics:
    contracts: list[Contract] = field(default_factory=list)
    abstractness: float = 0.0
    contract_metrics: OrderedDict = field(default_factory=OrderedDict)
    title: str = "Martin complexity metrics"
    full_text: str = ""
    core: SectionInfo = field(default=SectionInfo)
    CORE_KEYS = (
        "Dependents",
        "Dependencies",
        "Instability",
        "Distance from main sequence",
    )
    SECTIONS: tuple[tuple[str, str, tuple[str]]] = (("Core", "core", CORE_KEYS),)

    def __post_init__(self) -> None:
        self.update_abstractness()
        self.update_coupling()
        self.update_reporting_sections()

    def update_reporting_sections(self) -> None:
        # Create the table and text for each section.
        data = {
            contract.name: self.contract_metrics[contract.name].to_dict()
            for contract in self.contracts
        }
        for title, attr, keys in self.SECTIONS:
            pretty_table = make_pretty_table(["Contract", *keys], data, False)
            section_title = f"{self.title} ({title})"
            txt = f"\n\n{section_title}:\n"
            txt = "Martin agile software metrics\n"
            txt += "Efferent Coupling (Ce) - Number of contracts that a contract depends on\n"
            txt += "Afferent Coupling (Ca) - Number of contracts that depend on the contract\n"
            txt += (
                "Instability (I) - Ratio of efferent coupling to total coupling (Ce / (Ce + Ca))\n"
            )
            txt += "Abstractness (A) - Number of abstract contracts / total number of contracts\n"
            txt += "Distance from the Main Sequence (D) - abs(A + I - 1)\n"
            txt += "\n"
            txt += f"Abstractness (overall): {round(self.abstractness, 2)}\n"
            txt += f"{pretty_table}\n"
            self.full_text += txt
            setattr(
                self,
                attr,
                SectionInfo(title=section_title, pretty_table=pretty_table, txt=txt),
            )

    def update_abstractness(self) -> None:
        abstract_contract_count = 0
        for c in self.contracts:
            if not c.is_fully_implemented:
                abstract_contract_count += 1
        self.abstractness = float(abstract_contract_count / len(self.contracts))

    def update_coupling(self) -> None:
        dependencies = {}
        for contract in self.contracts:
            external_calls = []
            for func in contract.functions:
                high_level_calls = [
                    ir
                    for node in func.nodes
                    for ir in node.irs_ssa
                    if isinstance(ir, HighLevelCall)
                ]
                # convert irs to string with target function and contract name
                # Get the target contract name for each high level call
                new_external_calls = []
                for high_level_call in high_level_calls:
                    if isinstance(high_level_call.destination, Contract):
                        new_external_call = high_level_call.destination.name
                    elif isinstance(high_level_call.destination, str):
                        new_external_call = high_level_call.destination
                    elif not hasattr(high_level_call.destination, "type"):
                        continue
                    elif isinstance(high_level_call.destination.type, Contract):
                        new_external_call = high_level_call.destination.type.name
                    elif isinstance(high_level_call.destination.type, str):
                        new_external_call = high_level_call.destination.type
                    elif not hasattr(high_level_call.destination.type, "type"):
                        continue
                    elif isinstance(high_level_call.destination.type.type, Contract):
                        new_external_call = high_level_call.destination.type.type.name
                    elif isinstance(high_level_call.destination.type.type, str):
                        new_external_call = high_level_call.destination.type.type
                    else:
                        continue
                    new_external_calls.append(new_external_call)
                external_calls.extend(new_external_calls)
            dependencies[contract.name] = set(external_calls)
        dependents = {}
        for contract, deps in dependencies.items():
            for dep in deps:
                if dep not in dependents:
                    dependents[dep] = set()
                dependents[dep].add(contract)

        for contract in self.contracts:
            ce = len(dependencies.get(contract.name, []))
            ca = len(dependents.get(contract.name, []))
            self.contract_metrics[contract.name] = MartinContractMetrics(
                contract, ca, ce, self.abstractness
            )



from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue
from slither.slithir.variables.tuple import TupleVariable
from slither.core.variables.local_variable_init_from_tuple import LocalVariableInitFromTuple
from slither.slithir.variables.local_variable import LocalIRVariable
from slither.slithir.variables.tuple_ssa import TupleVariableSSA


class Unpack(OperationWithLValue):
    def __init__(
        self,
        result: LocalVariableInitFromTuple | LocalIRVariable,
        tuple_var: TupleVariable | TupleVariableSSA,
        idx: int,
    ) -> None:
        assert is_valid_lvalue(result)
        assert isinstance(tuple_var, TupleVariable)
        assert isinstance(idx, int)
        super().__init__()
        self._tuple = tuple_var
        self._idx = idx
        self._lvalue = result

    @property
    def read(self) -> list[TupleVariableSSA | TupleVariable]:
        return [self.tuple]

    @property
    def tuple(self) -> TupleVariable | TupleVariableSSA:
        return self._tuple

    @property
    def index(self) -> int:
        return self._idx

    def __str__(self):
        return f"{self.lvalue}({self.lvalue.type})= UNPACK {self.tuple} index: {self.index} "



from slither import Slither

# Init slither
slither = Slither("coin.sol")

for contract in slither.contracts:
    # Print the contract's name
    print(f"Contract: {contract.name}")
    # Print the name of the contract inherited
    print(f"\tInherit from{[c.name for c in contract.inheritance]}")
    for function in contract.functions:
        # For each function, print basic information
        print(f"\t{function.full_name}:")
        print(f"\t\tVisibility: {function.visibility}")
        print(f"\t\tContract: {function.contract}")
        print(f"\t\tModifier: {[m.name for m in function.modifiers]}")
        print(f"\t\tIs constructor? {function.is_constructor}")



"""
Re-entrancy detection

Based on heuristics, it may lead to FP and FN
Iterate over all the nodes of the graph until reaching a fixpoint
"""

from collections import namedtuple, defaultdict

from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.reentrancy.reentrancy import Reentrancy
from slither.utils.output import Output
from slither.slithir.operations import Binary, SolidityCall, HighLevelCall
from slither.core.declarations.function import Function
from slither.core.cfg.node import Node
from slither.analyses.data_dependency.data_dependency import is_dependent

# function: the vulnerable function
# call_node: the external call allowing reentrancy (may be internal function containing external call)
# external_calls: tuple of actual external call nodes (when call_node is internal function)
# balance_node: where balanceOf was called (before the external call)
FindingKey = namedtuple("FindingKey", ["function", "call_node", "external_calls", "balance_node"])
# variable: the variable in the guard that depends on stale balanceOf
# guard_node: the guard node using the stale value (after the external call)
FindingValue = namedtuple("FindingValue", ["variable", "guard_node"])


class ReentrancyBalance(Reentrancy):
    ARGUMENT = "reentrancy-balance"
    HELP = "Reentrancy vulnerabilities leading to outdated balance checks"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#reentrancy-vulnerabilities"
    )

    WIKI_TITLE = "Reentrancy vulnerabilities"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detects reentrancy vulnerabilities where a balance is saved (e.g., `balanceOf`) before an external call,
and the same balance is checked again after the call. An attacker could manipulate the balance during the reentrant call,
causing the post-call check to use an outdated value."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
interface IERC20 {
   function balanceOf(address account) external view returns (uint256);
}

interface I {
  function pay(uint256 amount) external;
}

function mint(IERC20 tk) public {
    uint amount_to_pay = 100;
    uint balance_before = tk.balanceOf(address(this));
    I(msg.sender).pay(amount_to_pay);
    require(tk.balanceOf(address(this)) - balance_before >= amount_to_pay);
    // Mint liquidity
}
```

The `balanceBefore` variable could be outdated if the external call reenter the `mint` function N times, allowing to pay the funds only once."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """Use `transferFrom` to transfer funds of a standard ERC20 token, or use reentrancy guards to prevent reentering the function."""

    STANDARD_JSON = False

    def _is_guard_node(self, node: Node) -> bool:
        """Check if node is a guard (require/assert or IF leading to revert)."""
        # Direct require/assert
        if node.contains_require_or_assert():
            return True

        # IF node - check if any immediate son is a revert
        if node.contains_if():
            for son in node.sons:
                # Check for revert() call in the son node
                for ir in son.irs:
                    if isinstance(ir, SolidityCall) and ir.function.name.startswith("revert"):
                        return True

        return False

    def find_reentrancies(self) -> defaultdict[FindingKey, set[FindingValue]]:
        result: defaultdict[FindingKey, set[FindingValue]] = defaultdict(set)
        for contract in self.contracts:
            for f in contract.functions_and_modifiers_declared:
                for node in f.nodes:
                    # Skip if no context or not a guard node
                    if self.KEY not in node.context:
                        continue
                    if not node.context[self.KEY].high_level_custom_calls_prior_calls:
                        continue
                    if not self._is_guard_node(node):
                        continue

                    # Check for external calls that could allow reentrancy
                    for call_node, external_call_nodes in node.context[self.KEY].calls.items():
                        prior_calls = node.context[
                            self.KEY
                        ].high_level_custom_calls_prior_calls.get(call_node, set())
                        for prior_ir in prior_calls:
                            found = False
                            for ir in node.irs:
                                if found:
                                    break
                                if isinstance(ir, Binary):
                                    for v in ir.read:
                                        if is_dependent(v, prior_ir.lvalue, node):
                                            # Sort external calls by node_id for deterministic output
                                            sorted_ext_calls = tuple(
                                                sorted(external_call_nodes, key=lambda x: x.node_id)
                                            )
                                            finding_key = FindingKey(
                                                function=f,
                                                call_node=call_node,
                                                external_calls=sorted_ext_calls,
                                                balance_node=prior_ir.node,
                                            )
                                            finding_value = FindingValue(
                                                variable=v,
                                                guard_node=node,
                                            )
                                            result[finding_key].add(finding_value)
                                            found = True
                                            break
                                    if found:
                                        break
        return result

    @staticmethod
    def custom_read_high_level_call(ir: HighLevelCall) -> bool:
        if isinstance(ir.function, Function) and ir.function.full_name == "balanceOf(address)":
            return True

        return False

    def _detect(self) -> list[Output]:
        super()._detect()
        reentrancies = self.find_reentrancies()

        results = []

        result_sorted = sorted(
            reentrancies.items(),
            key=lambda x: (x[0].function.name, x[0].call_node.node_id),
        )
        for (func, call_node, external_calls, balance_node), findings in result_sorted:
            findings = sorted(findings, key=lambda x: (x.variable.name, x.guard_node.node_id))

            info: list = ["Reentrancy in ", func, ":\n"]

            info += ["\tExternal call allowing reentrancy:\n"]
            info += ["\t- ", call_node, "\n"]
            # Show actual external calls if different from the call_node
            for ext_call in external_calls:
                if ext_call != call_node:
                    info += ["\t\t- ", ext_call, "\n"]

            info += ["\tBalance read before the call:\n"]
            info += ["\t- ", balance_node, "\n"]

            info += ["\tPossible stale balance used after the call in a condition:\n"]
            for finding in findings:
                info += ["\t- ", finding.guard_node, "\n"]
                info += ["\t\t- stale variable `", finding.variable.name, "`\n"]

            # Create our JSON result
            res = self.generate_result(info)

            # Add the function with the re-entrancy first
            res.add(func)

            # Add the external call
            res.add(call_node, {"underlying_type": "external_call"})
            # Add actual external calls if different from call_node
            for ext_call in external_calls:
                if ext_call != call_node:
                    res.add(ext_call, {"underlying_type": "external_call_actual"})

            # Add the balance read node (taint source)
            res.add(balance_node, {"underlying_type": "balance_before_call"})

            # Add guard nodes using stale balance
            for finding in findings:
                res.add(
                    finding.guard_node,
                    {
                        "underlying_type": "stale_balance_use",
                        "variable_name": finding.variable.name,
                    },
                )

            results.append(res)

        return results



import argparse
import logging
import sys

from crytic_compile import cryticparser
from crytic_compile.utils.zip import ZIP_TYPES_ACCEPTED

from slither import Slither
from slither.tools.flattening.flattening import (
    Flattening,
    Strategy,
    STRATEGIES_NAMES,
    DEFAULT_EXPORT_PATH,
)

logging.basicConfig()
logger = logging.getLogger("Slither")
logger.setLevel(logging.INFO)


def parse_args() -> argparse.Namespace:
    """
    Parse the underlying arguments for the program.
    :return: Returns the arguments for the program.
    """
    parser = argparse.ArgumentParser(
        description="Contracts flattening. See https://github.com/crytic/slither/wiki/Contract-Flattening",
        usage="slither-flat filename",
    )

    parser.add_argument("filename", help="The filename of the contract or project to analyze.")

    parser.add_argument("--contract", help="Flatten one contract.", default=None)

    parser.add_argument(
        "--strategy",
        help=f"Flatenning strategy: {STRATEGIES_NAMES} (default: MostDerived).",
        default=Strategy.MostDerived.name,
    )

    group_export = parser.add_argument_group("Export options")

    group_export.add_argument(
        "--dir",
        help=f"Export directory (default: {DEFAULT_EXPORT_PATH}).",
        default=None,
    )

    group_export.add_argument(
        "--json",
        help='Export the results as a JSON file ("--json -" to export to stdout)',
        action="store",
        default=None,
    )

    parser.add_argument(
        "--zip",
        help="Export all the files to a zip file",
        action="store",
        default=None,
    )

    parser.add_argument(
        "--zip-type",
        help=f"Zip compression type. One of {','.join(ZIP_TYPES_ACCEPTED.keys())}. Default lzma",
        action="store",
        default=None,
    )

    group_patching = parser.add_argument_group("Patching options")

    group_patching.add_argument(
        "--convert-external", help="Convert external to public.", action="store_true"
    )

    group_patching.add_argument(
        "--convert-private",
        help="Convert private variables to internal.",
        action="store_true",
    )

    group_patching.add_argument(
        "--convert-library-to-internal",
        help="Convert external or public functions to internal in library.",
        action="store_true",
    )

    group_patching.add_argument(
        "--remove-assert", help="Remove call to assert().", action="store_true"
    )

    group_patching.add_argument(
        "--pragma-solidity",
        help="Set the solidity pragma with a given version.",
        action="store",
        default=None,
    )

    # Add default arguments from crytic-compile
    cryticparser.init(parser)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def main() -> None:
    args = parse_args()

    slither = Slither(args.filename, **vars(args))

    for compilation_unit in slither.compilation_units:
        flat = Flattening(
            compilation_unit,
            external_to_public=args.convert_external,
            remove_assert=args.remove_assert,
            convert_library_to_internal=args.convert_library_to_internal,
            private_to_internal=args.convert_private,
            export_path=args.dir,
            pragma_solidity=args.pragma_solidity,
        )

        try:
            strategy = Strategy[args.strategy]
        except KeyError:
            to_log = f"{args.strategy} is not a valid strategy, use: {STRATEGIES_NAMES} (default MostDerived)"
            logger.error(to_log)
            return
        flat.export(
            strategy=strategy,
            target=args.contract,
            json=args.json,
            zip=args.zip,
            zip_type=args.zip_type,
        )


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from slither.core.declarations import Event
from slither.core.declarations.top_level import TopLevel

if TYPE_CHECKING:
    from slither.core.scope.scope import FileScope


class EventTopLevel(Event, TopLevel):
    def __init__(self, scope: "FileScope") -> None:
        super().__init__()
        self.file_scope: FileScope = scope



"""
Module detecting state variables that could be declared as constant
"""

from packaging import version
from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.core.variables.variable import Variable

from slither.visitors.expression.export_values import ExportValues
from slither.core.declarations import Contract, Function
from slither.core.declarations.solidity_variables import SolidityFunction
from slither.core.variables.state_variable import StateVariable
from slither.core.expressions import CallExpression, NewContract


def _is_valid_type(v: StateVariable) -> bool:
    t = v.type
    if isinstance(t, ElementaryType):
        return True
    if isinstance(t, UserDefinedType) and isinstance(t.type, Contract):
        return True
    return False


def _valid_candidate(v: StateVariable) -> bool:
    return _is_valid_type(v)


def _is_constant_var(v: Variable) -> bool:
    if isinstance(v, StateVariable):
        return v.is_constant
    return False


# https://solidity.readthedocs.io/en/v0.5.2/contracts.html#constant-state-variables
valid_solidity_function = [
    SolidityFunction("keccak256()"),
    SolidityFunction("keccak256(bytes)"),
    SolidityFunction("sha256()"),
    SolidityFunction("sha256(bytes)"),
    SolidityFunction("ripemd160()"),
    SolidityFunction("ripemd160(bytes)"),
    SolidityFunction("ecrecover(bytes32,uint8,bytes32,bytes32)"),
    SolidityFunction("addmod(uint256,uint256,uint256)"),
    SolidityFunction("mulmod(uint256,uint256,uint256)"),
]


def _constant_initial_expression(v: Variable) -> bool:
    if not v.expression:
        return True

    # B b = new B(); b cannot be constant, so filter out and recommend it be immutable
    if isinstance(v.expression, CallExpression) and isinstance(v.expression.called, NewContract):
        return False

    export = ExportValues(v.expression)
    values = export.result()
    if not values:
        return True

    return all(val in valid_solidity_function or _is_constant_var(val) for val in values)


class UnchangedStateVariables:
    """
    Find state variables that could be declared as constant or immutable (not written after deployment).
    """

    def __init__(self, compilation_unit: SlitherCompilationUnit) -> None:
        self.compilation_unit = compilation_unit
        self._constant_candidates: list[StateVariable] = []
        self._immutable_candidates: list[StateVariable] = []

    @property
    def immutable_candidates(self) -> list[StateVariable]:
        """Return the immutable candidates"""
        return self._immutable_candidates

    @property
    def constant_candidates(self) -> list[StateVariable]:
        """Return the constant candidates"""
        return self._constant_candidates

    def detect(self) -> None:
        """Detect state variables that could be constant or immutable"""
        for c in self.compilation_unit.contracts_derived:
            if c.is_signature_only():
                continue
            variables = []
            functions = []

            variables.append(c.storage_variables_ordered)
            functions.append(c.all_functions_called)

            valid_candidates: set[StateVariable] = {
                item for sublist in variables for item in sublist if _valid_candidate(item)
            }

            all_functions: list[Function] = list(
                {item1 for sublist in functions for item1 in sublist if isinstance(item1, Function)}
            )

            variables_written = []
            constructor_variables_written = []
            variables_initialized = []
            for f in all_functions:
                if f.is_constructor_variables:
                    variables_initialized.extend(f.state_variables_written)
                elif f.is_constructor:
                    constructor_variables_written.extend(f.state_variables_written)
                else:
                    variables_written.extend(f.state_variables_written)

            for v in valid_candidates:
                if v not in variables_written:
                    if _constant_initial_expression(v) and v not in constructor_variables_written:
                        self.constant_candidates.append(v)

                    elif (
                        not v.type.is_dynamic
                        and version.parse(self.compilation_unit.solc_version)
                        >= version.parse("0.6.5")
                        and (v in constructor_variables_written or v in variables_initialized)
                    ):
                        self.immutable_candidates.append(v)



import argparse
import inspect
import json
import logging
import sys
from typing import Any
from collections.abc import Sequence

from crytic_compile import cryticparser


from slither import Slither
from slither.core.declarations import Contract
from slither.exceptions import SlitherException
from slither.utils.colors import red
from slither.utils.output import output_to_json
from slither.tools.upgradeability.checks import all_checks
from slither.tools.upgradeability.checks.abstract_checks import (
    AbstractCheck,
    CheckClassification,
)
from slither.tools.upgradeability.utils.command_line import (
    output_detectors_json,
    output_wiki,
    output_detectors,
    output_to_markdown,
)

logging.basicConfig()
logger: logging.Logger = logging.getLogger("Slither")
logger.setLevel(logging.INFO)


def parse_args(check_classes: list[type[AbstractCheck]]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Slither Upgradeability Checks. For usage information see https://github.com/crytic/slither/wiki/Upgradeability-Checks.",
        usage="slither-check-upgradeability contract.sol ContractName",
    )

    group_checks = parser.add_argument_group("Checks")

    parser.add_argument("contract.sol", help="Codebase to analyze")
    parser.add_argument("ContractName", help="Contract name (logic contract)")

    parser.add_argument("--proxy-name", help="Proxy name")
    parser.add_argument("--proxy-filename", help="Proxy filename (if different)")

    parser.add_argument("--new-contract-name", help="New contract name (if changed)")
    parser.add_argument(
        "--new-contract-filename", help="New implementation filename (if different)"
    )

    parser.add_argument(
        "--json",
        help='Export the results as a JSON file ("--json -" to export to stdout)',
        action="store",
        default=False,
    )

    group_checks.add_argument(
        "--detect",
        help="Comma-separated list of detectors, defaults to all, "
        f"available detectors: {', '.join(d.ARGUMENT for d in check_classes)}",
        action="store",
        dest="detectors_to_run",
        default="all",
    )

    group_checks.add_argument(
        "--list-detectors",
        help="List available detectors",
        action=ListDetectors,
        nargs=0,
        default=False,
    )

    group_checks.add_argument(
        "--exclude",
        help="Comma-separated list of detectors that should be excluded",
        action="store",
        dest="detectors_to_exclude",
        default=None,
    )

    group_checks.add_argument(
        "--exclude-informational",
        help="Exclude informational impact analyses",
        action="store_true",
        default=False,
    )

    group_checks.add_argument(
        "--exclude-low",
        help="Exclude low impact analyses",
        action="store_true",
        default=False,
    )

    group_checks.add_argument(
        "--exclude-medium",
        help="Exclude medium impact analyses",
        action="store_true",
        default=False,
    )

    group_checks.add_argument(
        "--exclude-high",
        help="Exclude high impact analyses",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--markdown-root",
        help="URL for markdown generation",
        action="store",
        default="",
    )

    parser.add_argument(
        "--wiki-detectors", help=argparse.SUPPRESS, action=OutputWiki, default=False
    )

    parser.add_argument(
        "--list-detectors-json",
        help=argparse.SUPPRESS,
        action=ListDetectorsJson,
        nargs=0,
        default=False,
    )

    parser.add_argument("--markdown", help=argparse.SUPPRESS, action=OutputMarkdown, default=False)

    cryticparser.init(parser)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


###################################################################################
###################################################################################
# region checks
###################################################################################
###################################################################################


def _get_checks() -> list[type[AbstractCheck]]:
    detectors_ = [getattr(all_checks, name) for name in dir(all_checks)]
    detectors: list[type[AbstractCheck]] = [
        c for c in detectors_ if inspect.isclass(c) and issubclass(c, AbstractCheck)
    ]
    return detectors


def choose_checks(
    args: argparse.Namespace, all_check_classes: list[type[AbstractCheck]]
) -> list[type[AbstractCheck]]:
    detectors_to_run = []
    detectors = {d.ARGUMENT: d for d in all_check_classes}

    if args.detectors_to_run == "all":
        detectors_to_run = all_check_classes
        if args.detectors_to_exclude:
            detectors_excluded = args.detectors_to_exclude.split(",")
            for detector in detectors:
                if detector in detectors_excluded:
                    detectors_to_run.remove(detectors[detector])
    else:
        for detector in args.detectors_to_run.split(","):
            if detector in detectors:
                detectors_to_run.append(detectors[detector])
            else:
                raise Exception(f"Error: {detector} is not a detector")
        detectors_to_run = sorted(detectors_to_run, key=lambda x: x.IMPACT)
        return detectors_to_run

    if args.exclude_informational:
        detectors_to_run = [
            d for d in detectors_to_run if d.IMPACT != CheckClassification.INFORMATIONAL
        ]
    if args.exclude_low:
        detectors_to_run = [d for d in detectors_to_run if d.IMPACT != CheckClassification.LOW]
    if args.exclude_medium:
        detectors_to_run = [d for d in detectors_to_run if d.IMPACT != CheckClassification.MEDIUM]
    if args.exclude_high:
        detectors_to_run = [d for d in detectors_to_run if d.IMPACT != CheckClassification.HIGH]

    # detectors_to_run = sorted(detectors_to_run, key=lambda x: x.IMPACT)
    return detectors_to_run


class ListDetectors(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        checks = _get_checks()
        output_detectors(checks)
        parser.exit()


class ListDetectorsJson(argparse.Action):
    def __call__(self, parser: Any, *args: Any, **kwargs: Any) -> None:
        checks = _get_checks()
        detector_types_json = output_detectors_json(checks)
        print(json.dumps(detector_types_json))
        parser.exit()


class OutputMarkdown(argparse.Action):
    def __call__(
        self,
        parser: Any,
        args: Any,
        values: str | Sequence[Any] | None,
        option_string: Any = None,
    ) -> None:
        checks = _get_checks()
        assert isinstance(values, str)
        output_to_markdown(checks, values)
        parser.exit()


class OutputWiki(argparse.Action):
    def __call__(
        self,
        parser: Any,
        args: Any,
        values: str | Sequence[Any] | None,
        option_string: Any = None,
    ) -> Any:
        checks = _get_checks()
        assert isinstance(values, str)
        output_wiki(checks, values)
        parser.exit()


def _run_checks(detectors: list[AbstractCheck]) -> list[dict]:
    results_ = [d.check() for d in detectors]
    results_ = [r for r in results_ if r]
    results = [item for sublist in results_ for item in sublist]  # flatten
    return results


def _checks_on_contract(
    detectors: list[type[AbstractCheck]], contract: Contract
) -> tuple[list[dict], int]:
    detectors_ = [
        d(logger, contract)
        for d in detectors
        if (not d.REQUIRE_PROXY and not d.REQUIRE_CONTRACT_V2)
    ]
    return _run_checks(detectors_), len(detectors_)


def _checks_on_contract_update(
    detectors: list[type[AbstractCheck]], contract_v1: Contract, contract_v2: Contract
) -> tuple[list[dict], int]:
    detectors_ = [
        d(logger, contract_v1, contract_v2=contract_v2) for d in detectors if d.REQUIRE_CONTRACT_V2
    ]
    return _run_checks(detectors_), len(detectors_)


def _checks_on_contract_and_proxy(
    detectors: list[type[AbstractCheck]], contract: Contract, proxy: Contract
) -> tuple[list[dict], int]:
    detectors_ = [d(logger, contract, proxy=proxy) for d in detectors if d.REQUIRE_PROXY]
    return _run_checks(detectors_), len(detectors_)


# endregion
###################################################################################
###################################################################################
# region Main
###################################################################################
###################################################################################


def main() -> None:
    json_results: dict = {
        "proxy-present": False,
        "contract_v2-present": False,
        "detectors": [],
    }

    detectors = _get_checks()
    args = parse_args(detectors)
    detectors_to_run = choose_checks(args, detectors)
    v1_filename = vars(args)["contract.sol"]
    number_detectors_run = 0
    try:
        variable1 = Slither(v1_filename, **vars(args))

        # Analyze logic contract
        v1_name = args.ContractName
        v1_contracts = variable1.get_contract_from_name(v1_name)
        if len(v1_contracts) != 1:
            info = f"Contract {v1_name} not found in {variable1.filename}"
            logger.error(red(info))
            if args.json:
                output_to_json(args.json, str(info), json_results)
            return
        v1_contract = v1_contracts[0]

        detectors_results, number_detectors = _checks_on_contract(detectors_to_run, v1_contract)
        json_results["detectors"] += detectors_results
        number_detectors_run += number_detectors

        # Analyze Proxy
        proxy_contract = None
        if args.proxy_name:
            if args.proxy_filename:
                proxy = Slither(args.proxy_filename, **vars(args))
            else:
                proxy = variable1

            proxy_contracts = proxy.get_contract_from_name(args.proxy_name)
            if len(proxy_contracts) != 1:
                info = f"Proxy {args.proxy_name} not found in {proxy.filename}"
                logger.error(red(info))
                if args.json:
                    output_to_json(args.json, str(info), json_results)
                return
            proxy_contract = proxy_contracts[0]
            json_results["proxy-present"] = True

            detectors_results, number_detectors = _checks_on_contract_and_proxy(
                detectors_to_run, v1_contract, proxy_contract
            )
            json_results["detectors"] += detectors_results
            number_detectors_run += number_detectors
        # Analyze new version
        if args.new_contract_name:
            if args.new_contract_filename:
                variable2 = Slither(args.new_contract_filename, **vars(args))
            else:
                variable2 = variable1

            v2_contracts = variable2.get_contract_from_name(args.new_contract_name)
            if len(v2_contracts) != 1:
                info = (
                    f"New logic contract {args.new_contract_name} not found in {variable2.filename}"
                )
                logger.error(red(info))
                if args.json:
                    output_to_json(args.json, str(info), json_results)
                return
            v2_contract = v2_contracts[0]
            json_results["contract_v2-present"] = True

            if proxy_contract:
                detectors_results, _ = _checks_on_contract_and_proxy(
                    detectors_to_run, v2_contract, proxy_contract
                )

                json_results["detectors"] += detectors_results

            detectors_results, number_detectors = _checks_on_contract_update(
                detectors_to_run, v1_contract, v2_contract
            )
            json_results["detectors"] += detectors_results
            number_detectors_run += number_detectors

            # If there is a V2, we run the contract-only check on the V2
            detectors_results, number_detectors = _checks_on_contract(detectors_to_run, v2_contract)
            json_results["detectors"] += detectors_results
            number_detectors_run += number_detectors

        to_log = f"{len(json_results['detectors'])} findings, {number_detectors_run} detectors run"
        logger.info(to_log)
        if args.json:
            output_to_json(args.json, None, json_results)

    except SlitherException as slither_exception:
        logger.error(str(slither_exception))
        if args.json:
            output_to_json(args.json, str(slither_exception), json_results)
        return


# endregion



from pathlib import Path
from slither import Slither

from slither.core.expressions import CallExpression
from slither.core.expressions.identifier import Identifier
from slither.core.expressions.literal import Literal
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.local_variable import LocalVariable


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_yul_parser_assembly_slot(solc_binary_path) -> None:
    # mstore(0x0, bucketId)
    # mstore(0x20, _counters.slot)
    data = {"0x0": "bucketId", "0x20": "_counters"}

    solc_path = solc_binary_path("0.8.18")
    slither = Slither(Path(TEST_DATA_DIR, "assembly_storage_slot.sol").as_posix(), solc=solc_path)

    contract = slither.get_contract_from_name("XXX")[0]
    func = contract.get_function_from_full_name("getPackedBucketGlobalState(uint256)")
    calls = [
        node.expression
        for node in func.all_nodes()
        if node.expression and "mstore" in str(node.expression)
    ]

    for call in calls:
        assert isinstance(call, CallExpression)
        memory_location = call.arguments[0]
        value = call.arguments[1]
        assert isinstance(memory_location, Literal)
        assert isinstance(value, Identifier)
        assert value.value.name == data[memory_location.value]
        if value.value.name == "_counters":
            assert isinstance(value.value, StateVariable)
        elif value.value.name == "bucketId":
            assert isinstance(value.value, LocalVariable)


def test_yul_parser_sstore_sload(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.18")
    slither = Slither(Path(TEST_DATA_DIR, "assembly_sstore_sload.sol").as_posix(), solc=solc_path)

    contract = slither.get_contract_from_name("Test")[0]

    read = contract.get_function_from_full_name("read()")
    # Sload is converted to an assignement
    assert len(read.all_solidity_calls()) == 0

    read_parameter = contract.get_function_from_full_name("read_parameter(uint256)")
    # Sload is kept because the slot is dynamic
    assert len(read_parameter.all_solidity_calls()) == 1

    write = contract.get_function_from_full_name("write()")
    # Sstore is converted to an assignement
    assert len(write.all_solidity_calls()) == 0

    write_parameter = contract.get_function_from_full_name("write_parameter(uint256)")
    # Sstore is kept because the slot is dynamic
    assert len(write_parameter.all_solidity_calls()) == 1



from pathlib import Path
from slither import Slither
from slither.core.cfg.node import NodeType
from slither.slithir.operations import Assignment, Unpack
from slither.core.expressions import (
    AssignmentOperation,
    TupleExpression,
    NewElementaryType,
    CallExpression,
)


TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_ternary_conversions(solc_binary_path) -> None:
    """This tests that true and false sons define the same number of variables that the father node declares"""
    solc_path = solc_binary_path("0.8.0")
    slither = Slither(Path(TEST_DATA_DIR, "ternary_expressions.sol").as_posix(), solc=solc_path)
    contract = next(c for c in slither.contracts if c.name == "C")
    for function in contract.functions:
        vars_declared = 0
        vars_assigned = 0
        for node in function.nodes:
            if node.type in [NodeType.IF, NodeType.IFLOOP]:
                # Iterate over true and false son
                for inner_node in node.sons:
                    # Count all variables declared
                    expression = inner_node.expression
                    if isinstance(
                        expression, (AssignmentOperation, NewElementaryType, CallExpression)
                    ):
                        var_expr = expression.expression_left
                        # Only tuples declare more than one var
                        if isinstance(var_expr, TupleExpression):
                            for expr in var_expr.expressions:
                                if expr is not None:
                                    vars_declared += 1
                        else:
                            vars_declared += 1

                    for ir in inner_node.irs:
                        # Count all variables defined
                        if isinstance(ir, (Assignment, Unpack)):
                            vars_assigned += 1
        assert vars_declared == vars_assigned and vars_assigned != 0


def test_ternary_tuple(solc_binary_path) -> None:
    """
    Test that in the ternary liftings of an assignment of the form `(z, ) = ...`,
    we obtain `z` from an unpack operation in both lifitings
    """
    solc_path = solc_binary_path("0.8.0")
    slither = Slither(Path(TEST_DATA_DIR, "ternary_expressions.sol").as_posix(), solc=solc_path)
    contract = next(c for c in slither.contracts if c.name == "D")
    fn = next(f for f in contract.functions if f.name == "a")

    if_nodes = [n for n in fn.nodes if n.type == NodeType.IF]
    assert len(if_nodes) == 1

    if_node = if_nodes[0]
    assert isinstance(if_node.son_true.expression, AssignmentOperation)
    assert (
        len([ir for ir in if_node.son_true.all_slithir_operations() if isinstance(ir, Unpack)]) == 1
    )
    assert (
        len([ir for ir in if_node.son_false.all_slithir_operations() if isinstance(ir, Unpack)])
        == 1
    )



"""
Module detecting uninitialized function pointer calls in constructors
"""

from typing import Any
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    make_solc_versions,
    DETECTOR_INFO,
)
from slither.slithir.operations import InternalDynamicCall, OperationWithLValue
from slither.slithir.variables import ReferenceVariable
from slither.slithir.variables.variable import SlithIRVariable
from slither.core.cfg.node import Node
from slither.core.declarations.contract import Contract
from slither.core.declarations.function_contract import FunctionContract
from slither.slithir.variables.state_variable import StateIRVariable
from slither.utils.output import Output


def _get_variables_entrance(function: FunctionContract) -> list[Any | StateIRVariable]:
    """
    Return the first SSA variables of the function
    Catpure the phi operation at the entry point
    """
    ret = []
    if function.entry_point:
        for ir_ssa in function.entry_point.irs_ssa:
            if isinstance(ir_ssa, OperationWithLValue):
                ret.append(ir_ssa.lvalue)
    return ret


def _is_vulnerable(node: Node, variables_entrance: list[Any | StateIRVariable]) -> bool:
    """
    Vulnerable if an IR ssa:
        - It is an internal dynamic call
        - The destination has not an index of 0
        - The destination is not in the allowed variable
    """
    for ir_ssa in node.irs_ssa:
        if isinstance(ir_ssa, InternalDynamicCall):
            destination = ir_ssa.function
            # If it is a reference variable, destination should be the origin variable
            # Note: this will create FN if one of the field of a structure is updated, while not begin
            # the field of the function pointer. This should be fixed once we have the IR refactoring
            if isinstance(destination, ReferenceVariable):
                destination = destination.points_to_origin
            if isinstance(destination, SlithIRVariable) and ir_ssa.function.index == 0:
                return True
            if destination in variables_entrance:
                return True
    return False


class UninitializedFunctionPtrsConstructor(AbstractDetector):
    """
    Uninitialized function pointer calls in constructors
    """

    ARGUMENT = "uninitialized-fptr-cst"
    HELP = "Uninitialized function pointer calls in constructors"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#uninitialized-function-pointers-in-constructors"
    WIKI_TITLE = "Uninitialized function pointers in constructors"
    WIKI_DESCRIPTION = "solc versions `0.4.5`-`0.4.26` and `0.5.0`-`0.5.8` contain a compiler bug leading to unexpected behavior when calling uninitialized function pointers in constructors."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract bad0 {

  constructor() public {
    /* Uninitialized function pointer */
    function(uint256) internal returns(uint256) a;
    a(10);
  }

}
```
The call to `a(10)` will lead to unexpected behavior because function pointer `a` is not initialized in the constructor."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = (
        "Initialize function pointers before calling. Avoid function pointers if possible."
    )

    VULNERABLE_SOLC_VERSIONS = make_solc_versions(4, 5, 25) + make_solc_versions(5, 0, 8)

    @staticmethod
    def _detect_uninitialized_function_ptr_in_constructor(
        contract: Contract,
    ) -> list[Any | Node]:
        """
        Detect uninitialized function pointer calls in constructors
        :param contract: The contract of interest for detection
        :return: A list of nodes with uninitialized function pointer calls in the constructor of given contract
        """
        results = []
        constructor = contract.constructors_declared
        if constructor:
            variables_entrance = _get_variables_entrance(constructor)
            results = [
                node for node in constructor.nodes if _is_vulnerable(node, variables_entrance)
            ]
        return results

    def _detect(self) -> list[Output]:
        """
        Detect uninitialized function pointer calls in constructors of contracts
        Returns:
            list: ['uninitialized function pointer calls in constructors']
        """
        results = []

        for contract in self.compilation_unit.contracts:
            contract_info: DETECTOR_INFO = ["Contract ", contract, " \n"]
            nodes = self._detect_uninitialized_function_ptr_in_constructor(contract)
            for node in nodes:
                node_info: DETECTOR_INFO = [
                    "\t ",
                    node,
                    " is an uninitialized function pointer call in a constructor\n",
                ]
                json = self.generate_result(contract_info + node_info)
                results.append(json)

        return results



from __future__ import annotations
from dataclasses import dataclass


@dataclass(slots=True)
class ASTNode:
    src: str
    node_id: int


@dataclass(slots=True)
class Definition(ASTNode):
    doc_string: str | None


@dataclass(slots=True)
class Module(Definition):
    body: list[ASTNode]
    name: str


@dataclass(slots=True)
class ImportFrom(ASTNode):
    module: str
    name: str
    alias: str | None


@dataclass(slots=True)
class EventDef(ASTNode):
    name: str
    body: list[AnnAssign]


@dataclass(slots=True)
class AnnAssign(ASTNode):
    target: Name
    annotation: Subscript | Name | Call
    value: ASTNode | None


@dataclass(slots=True)
class Name(ASTNode):  # type or identifier
    id: str


@dataclass(slots=True)
class Call(ASTNode):
    func: ASTNode
    args: list[ASTNode]
    keyword: ASTNode | None
    keywords: list[ASTNode]


@dataclass(slots=True)
class Pass(ASTNode):
    pass


@dataclass(slots=True)
class StructDef(ASTNode):
    name: str
    body: list[AnnAssign]


@dataclass(slots=True)
class VariableDecl(ASTNode):
    annotation: ASTNode
    target: ASTNode
    value: ASTNode | None
    is_constant: bool
    is_immutable: bool
    is_public: bool


@dataclass(slots=True)
class Subscript(ASTNode):
    value: ASTNode
    slice: ASTNode


@dataclass(slots=True)
class Index(ASTNode):
    value: ASTNode


@dataclass(slots=True)
class Bytes(ASTNode):
    value: bytes


@dataclass(slots=True)
class Hex(ASTNode):
    value: str


@dataclass(slots=True)
class Int(ASTNode):
    value: int


@dataclass(slots=True)
class Str(ASTNode):
    value: str


@dataclass(slots=True)
class VyList(ASTNode):
    elements: list[ASTNode]


@dataclass(slots=True)
class VyDict(ASTNode):
    keys: list[ASTNode]
    values: list[ASTNode]


@dataclass(slots=True)
class Tuple(ASTNode):
    elements: list[ASTNode]


@dataclass(slots=True)
class FunctionDef(Definition):
    name: str
    args: Arguments | None
    returns: list[ASTNode] | None
    body: list[ASTNode]
    decorators: list[ASTNode] | None
    pos: any | None  # not sure what this is


@dataclass(slots=True)
class Assign(ASTNode):
    target: ASTNode
    value: ASTNode


@dataclass(slots=True)
class Attribute(ASTNode):
    value: ASTNode
    attr: str


@dataclass(slots=True)
class Arguments(ASTNode):
    args: list[Arg]
    default: ASTNode | None
    defaults: list[ASTNode]


@dataclass(slots=True)
class Arg(ASTNode):
    arg: str
    annotation: ASTNode | None


@dataclass(slots=True)
class Assert(ASTNode):
    test: ASTNode
    msg: Str | None


@dataclass(slots=True)
class Raise(ASTNode):
    exc: ASTNode


@dataclass(slots=True)
class Expr(ASTNode):
    value: ASTNode


@dataclass(slots=True)
class UnaryOp(ASTNode):
    op: ASTNode
    operand: ASTNode


@dataclass(slots=True)
class BinOp(ASTNode):
    left: ASTNode
    op: str
    right: ASTNode


@dataclass(slots=True)
class Keyword(ASTNode):
    arg: str
    value: ASTNode


@dataclass(slots=True)
class Log(ASTNode):
    value: ASTNode


@dataclass(slots=True)
class Return(ASTNode):
    value: ASTNode | None


@dataclass(slots=True)
class If(ASTNode):
    test: ASTNode
    body: list[ASTNode]
    orelse: list[ASTNode]


@dataclass(slots=True)
class Compare(ASTNode):
    left: ASTNode
    op: ASTNode
    right: ASTNode


@dataclass(slots=True)
class NameConstant(ASTNode):
    value: bool


@dataclass(slots=True)
class For(ASTNode):
    target: ASTNode
    iter: ASTNode
    body: list[ASTNode]


@dataclass(slots=True)
class Continue(ASTNode):
    pass


@dataclass(slots=True)
class Break(ASTNode):
    pass


@dataclass(slots=True)
class InterfaceDef(ASTNode):
    name: str
    body: list[ASTNode]


@dataclass(slots=True)
class EnumDef(ASTNode):
    name: str
    body: list[ASTNode]


@dataclass(slots=True)
class AugAssign(ASTNode):
    target: ASTNode
    op: ASTNode
    value: ASTNode


@dataclass(slots=True)
class BoolOp(ASTNode):
    op: ASTNode
    values: list[ASTNode]



"""
We use protected member, to avoid having setter in the expression
as they should be immutable
"""

import copy
from collections.abc import Callable

from slither.all_exceptions import SlitherException
from slither.core.expressions import UnaryOperation
from slither.core.expressions.assignment_operation import AssignmentOperation
from slither.core.expressions.binary_operation import BinaryOperation
from slither.core.expressions.call_expression import CallExpression
from slither.core.expressions.conditional_expression import ConditionalExpression
from slither.core.expressions.elementary_type_name_expression import ElementaryTypeNameExpression
from slither.core.expressions.expression import Expression
from slither.core.expressions.identifier import Identifier
from slither.core.expressions.index_access import IndexAccess
from slither.core.expressions.literal import Literal
from slither.core.expressions.member_access import MemberAccess
from slither.core.expressions.new_array import NewArray
from slither.core.expressions.new_contract import NewContract
from slither.core.expressions.tuple_expression import TupleExpression
from slither.core.expressions.type_conversion import TypeConversion
from slither.core.expressions.new_elementary_type import NewElementaryType


def f_expressions(
    e: AssignmentOperation | BinaryOperation | TupleExpression,
    x: Identifier | Literal | MemberAccess | IndexAccess,
) -> None:
    e._expressions.append(x)


def f_call(e: CallExpression, x: ElementaryTypeNameExpression) -> None:
    e._arguments.append(x)


def f_call_value(e: CallExpression, x):
    e._value = x


def f_call_gas(e: CallExpression, x):
    e._gas = x


def f_expression(e: TypeConversion | UnaryOperation | MemberAccess, x: CallExpression) -> None:
    e._expression = x


def f_called(e: CallExpression, x: Identifier) -> None:
    e._called = x


class SplitTernaryExpression:
    def __init__(self, expression: AssignmentOperation | ConditionalExpression) -> None:
        if isinstance(expression, ConditionalExpression):
            self.true_expression = copy.copy(expression.then_expression)
            self.false_expression = copy.copy(expression.else_expression)
            self.condition = copy.copy(expression.if_expression)
        else:
            self.true_expression = copy.copy(expression)
            self.false_expression = copy.copy(expression)
            self.condition = None
            self.copy_expression(expression, self.true_expression, self.false_expression)

    def conditional_not_ahead(
        self,
        next_expr: Expression,
        true_expression: AssignmentOperation | MemberAccess,
        false_expression: AssignmentOperation | MemberAccess,
        f: Callable,
    ) -> bool:
        # look ahead for parenthetical expression (.. ? .. : ..)
        if (
            isinstance(next_expr, TupleExpression)
            and len(next_expr.expressions) == 1
            and isinstance(next_expr.expressions[0], ConditionalExpression)
        ):
            next_expr = next_expr.expressions[0]

        if isinstance(next_expr, ConditionalExpression):
            f(true_expression, copy.copy(next_expr.then_expression))
            f(false_expression, copy.copy(next_expr.else_expression))
            self.condition = copy.copy(next_expr.if_expression)
            return False

        f(true_expression, copy.copy(next_expr))
        f(false_expression, copy.copy(next_expr))
        return True

    def copy_expression(
        self, expression: Expression, true_expression: Expression, false_expression: Expression
    ) -> None:
        if self.condition:
            return

        if isinstance(expression, ConditionalExpression):
            raise SlitherException("Nested ternary operator not handled")

        if isinstance(
            expression,
            (
                Literal,
                Identifier,
                NewArray,
                NewContract,
                ElementaryTypeNameExpression,
                NewElementaryType,
            ),
        ):
            return

        if isinstance(
            expression, (AssignmentOperation, BinaryOperation, TupleExpression, IndexAccess)
        ):
            true_expression._expressions = []
            false_expression._expressions = []
            self.convert_expressions(expression, true_expression, false_expression)

        elif isinstance(expression, CallExpression):
            next_expr = expression.called
            self.convert_call_expression(expression, next_expr, true_expression, false_expression)

        elif isinstance(expression, (TypeConversion, UnaryOperation, MemberAccess)):
            next_expr = expression.expression
            if self.conditional_not_ahead(
                next_expr, true_expression, false_expression, f_expression
            ):
                self.copy_expression(
                    expression.expression,
                    true_expression.expression,
                    false_expression.expression,
                )

        else:
            raise SlitherException(
                f"Ternary operation not handled {expression}({type(expression)})"
            )

    def convert_expressions(
        self,
        expression: AssignmentOperation | BinaryOperation | TupleExpression,
        true_expression: Expression,
        false_expression: Expression,
    ) -> None:
        for next_expr in expression.expressions:
            # TODO: can we get rid of `NoneType` expressions in `TupleExpression`?
            # montyly: this might happen with unnamed tuple (ex: (,,,) = f()), but it needs to be checked
            if next_expr is not None:
                if self.conditional_not_ahead(
                    next_expr, true_expression, false_expression, f_expressions
                ):
                    # always on last arguments added
                    self.copy_expression(
                        next_expr,
                        true_expression.expressions[-1],
                        false_expression.expressions[-1],
                    )
            else:
                true_expression.expressions.append(None)
                false_expression.expressions.append(None)

    def convert_index_access(
        self, next_expr: IndexAccess, true_expression: Expression, false_expression: Expression
    ) -> None:
        # create an index access for each branch
        #  x[if cond ? 1 : 2] -> if cond { x[1] } else { x[2] }
        for expr in next_expr.expressions:
            if self.conditional_not_ahead(expr, true_expression, false_expression, f_expressions):
                self.copy_expression(
                    expr,
                    true_expression.expressions[-1],
                    false_expression.expressions[-1],
                )

    def convert_call_expression(
        self,
        expression: CallExpression,
        next_expr: Expression,
        true_expression: Expression,
        false_expression: Expression,
    ) -> None:
        # case of lib
        # (.. ? .. : ..).add
        if self.conditional_not_ahead(next_expr, true_expression, false_expression, f_called):
            self.copy_expression(next_expr, true_expression.called, false_expression.called)

        # In order to handle ternaries in both call options, gas and value, we return early if the
        # conditional is not ahead to rewrite both ternaries (see `_rewrite_ternary_as_if_else`).
        if expression.call_gas:
            # case of (..).func{gas: .. ? .. : ..}()
            next_expr = expression.call_gas
            if self.conditional_not_ahead(next_expr, true_expression, false_expression, f_call_gas):
                self.copy_expression(
                    next_expr,
                    true_expression.call_gas,
                    false_expression.call_gas,
                )
            else:
                return

        if expression.call_value:
            # case of (..).func{value: .. ? .. : ..}()
            next_expr = expression.call_value
            if self.conditional_not_ahead(
                next_expr, true_expression, false_expression, f_call_value
            ):
                self.copy_expression(
                    next_expr,
                    true_expression.call_value,
                    false_expression.call_value,
                )
            else:
                return

        true_expression._arguments = []
        false_expression._arguments = []

        for expr in expression.arguments:
            if self.conditional_not_ahead(expr, true_expression, false_expression, f_call):
                # always on last arguments added
                self.copy_expression(
                    expr,
                    true_expression.arguments[-1],
                    false_expression.arguments[-1],
                )



from slither.core.expressions.expression import Expression
from slither.core.solidity_types.elementary_type import ElementaryType


class NewElementaryType(Expression):
    def __init__(self, new_type):
        assert isinstance(new_type, ElementaryType)
        super().__init__()
        self._type = new_type

    @property
    def type(self) -> ElementaryType:
        return self._type

    def __str__(self):
        return "new " + str(self._type)



import logging
import operator
import sys
import traceback
from argparse import Namespace

from slither.tools.similarity.encode import encode_contract, load_and_encode, parse_target
from slither.tools.similarity.model import load_model
from slither.tools.similarity.similarity import similarity

logger = logging.getLogger("Slither-simil")


def test(args: Namespace) -> None:
    try:
        model = args.model
        model = load_model(model)
        filename = args.filename
        contract, fname = parse_target(args.fname)
        infile = args.input
        ntop = args.ntop

        if filename is None or contract is None or fname is None or infile is None:
            logger.error("The test mode requires filename, contract, fname and input parameters.")
            sys.exit(-1)

        irs = encode_contract(filename, **vars(args))
        if len(irs) == 0:
            sys.exit(-1)

        y = " ".join(irs[(filename, contract, fname)])

        fvector = model.get_sentence_vector(y)
        cache = load_and_encode(infile, model, **vars(args))
        # save_cache("cache.npz", cache)

        r = {}
        for x, y in cache.items():
            r[x] = similarity(fvector, y)

        r = sorted(r.items(), key=operator.itemgetter(1), reverse=True)
        logger.info("Reviewed %d functions, listing the %d most similar ones:", len(r), ntop)
        format_table = "{: <65} {: <20} {: <20} {: <10}"
        logger.info(format_table.format(*["filename", "contract", "function", "score"]))
        for x, score in r[:ntop]:
            score = str(round(score, 3))
            logger.info(format_table.format(*(list(x) + [score])))

    except Exception:
        logger.error(f"Error in {args.filename}")
        logger.error(traceback.format_exc())
        sys.exit(-1)



"""
Module detecting redundant statements.
"""

from slither.core.cfg.node import Node, NodeType
from slither.core.declarations.contract import Contract
from slither.core.expressions.elementary_type_name_expression import ElementaryTypeNameExpression
from slither.core.expressions.identifier import Identifier
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class RedundantStatements(AbstractDetector):
    """
    Use of Redundant Statements
    """

    ARGUMENT = "redundant-statements"
    HELP = "Redundant statements"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#redundant-statements"

    WIKI_TITLE = "Redundant Statements"
    WIKI_DESCRIPTION = "Detect the usage of redundant statements that have no effect."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract RedundantStatementsContract {

    constructor() public {
        uint; // Elementary Type Name
        bool; // Elementary Type Name
        RedundantStatementsContract; // Identifier
    }

    function test() public returns (uint) {
        uint; // Elementary Type Name
        assert; // Identifier
        test; // Identifier
        return 777;
    }
}
```
Each commented line references types/identifiers, but performs no action with them, so no code will be generated for such statements and they can be removed."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Remove redundant statements if they congest code but offer no value."

    # This is a disallowed list of tuple (node.type, type(node.expression))
    REDUNDANT_TOP_LEVEL_EXPRESSIONS = (ElementaryTypeNameExpression, Identifier)

    def detect_redundant_statements_contract(self, contract: Contract) -> list[Node]:
        """Detects the usage of redundant statements in a contract.

        Returns:
            list: nodes"""
        results = []

        # Loop through all functions + modifiers defined explicitly in this contract.
        for function in contract.functions_and_modifiers_declared:
            # Loop through each node in this function.
            for node in function.nodes:
                if node.expression:
                    if node.type == NodeType.EXPRESSION and isinstance(
                        node.expression, self.REDUNDANT_TOP_LEVEL_EXPRESSIONS
                    ):
                        results.append(node)

        return results

    def _detect(self) -> list[Output]:
        """Detect redundant statements

        Recursively visit the calls
        Returns:
            list: {'vuln', 'filename,'contract','func', 'redundant_statements'}

        """
        results = []
        for contract in self.contracts:
            redundant_statements = self.detect_redundant_statements_contract(contract)
            if redundant_statements:
                for redundant_statement in redundant_statements:
                    info: DETECTOR_INFO = [
                        'Redundant expression "',
                        redundant_statement,
                        '" in',
                        contract,
                        "\n",
                    ]
                    json = self.generate_result(info)
                    results.append(json)

        return results



import logging
from pathlib import Path
from typing import Any, TYPE_CHECKING


from slither.core.declarations import (
    Contract,
    Enum,
    Event,
    Function,
    SolidityFunction,
    SolidityVariable,
    SolidityVariableComposed,
    Structure,
)
from slither.core.declarations.custom_error import CustomError
from slither.core.declarations.function_contract import FunctionContract
from slither.core.declarations.function_top_level import FunctionTopLevel
from slither.core.declarations.solidity_import_placeholder import SolidityImportPlaceHolder
from slither.core.declarations.solidity_variables import SolidityCustomRevert
from slither.core.expressions import Identifier, Literal
from slither.core.expressions.expression import Expression
from slither.core.solidity_types import (
    ArrayType,
    ElementaryType,
    FunctionType,
    MappingType,
    UserDefinedType,
    TypeInformation,
)
from slither.core.solidity_types.elementary_type import (
    Int as ElementaryTypeInt,
    Uint,
    Byte,
    MaxValues,
)
from slither.core.solidity_types.type import Type
from slither.core.solidity_types.type_alias import TypeAliasTopLevel, TypeAlias
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.variable import Variable
from slither.slithir.exceptions import SlithIRError
from slither.slithir.operations import (
    Assignment,
    Binary,
    BinaryType,
    Call,
    Condition,
    Delete,
    EventCall,
    HighLevelCall,
    Index,
    InitArray,
    InternalCall,
    InternalDynamicCall,
    Length,
    LibraryCall,
    LowLevelCall,
    Member,
    NewArray,
    NewContract,
    NewElementaryType,
    NewStructure,
    OperationWithLValue,
    Return,
    Send,
    SolidityCall,
    Transfer,
    TypeConversion,
    Unary,
    Unpack,
    Nop,
    Operation,
)
from slither.slithir.operations.codesize import CodeSize
from slither.slithir.tmp_operations.argument import Argument, ArgumentType
from slither.slithir.tmp_operations.tmp_call import TmpCall
from slither.slithir.tmp_operations.tmp_new_array import TmpNewArray
from slither.slithir.tmp_operations.tmp_new_contract import TmpNewContract
from slither.slithir.tmp_operations.tmp_new_elementary_type import TmpNewElementaryType
from slither.slithir.tmp_operations.tmp_new_structure import TmpNewStructure
from slither.slithir.variables import Constant, ReferenceVariable, TemporaryVariable
from slither.slithir.variables import TupleVariable
from slither.utils.function import get_function_id, get_event_id
from slither.utils.type import export_nested_types_from_variable
from slither.utils.using_for import USING_FOR
from slither.visitors.slithir.expression_to_slithir import ExpressionToSlithIR

if TYPE_CHECKING:
    from slither.core.cfg.node import Node

logger = logging.getLogger("ConvertToIR")


def convert_expression(expression: Expression, node: "Node") -> list[Operation]:
    # handle standlone expression
    # such as return true;
    from slither.core.cfg.node import NodeType

    if isinstance(expression, Literal) and node.type in [NodeType.IF, NodeType.IFLOOP]:
        cst = Constant(expression.value, expression.type)
        cond = Condition(cst)
        cond.set_expression(expression)
        cond.set_node(node)
        return [cond]
    if isinstance(expression, Identifier) and node.type in [
        NodeType.IF,
        NodeType.IFLOOP,
    ]:
        cond = Condition(expression.value)
        cond.set_expression(expression)
        cond.set_node(node)
        return [cond]

    visitor = ExpressionToSlithIR(expression, node)
    result = visitor.result()
    is_solidity = node.compilation_unit.is_solidity
    result = apply_ir_heuristics(result, node, is_solidity)

    if result:
        if node.type in [NodeType.IF, NodeType.IFLOOP]:
            prev = result[-1]
            assert isinstance(prev, (OperationWithLValue)) and prev.lvalue
            cond = Condition(prev.lvalue)
            cond.set_expression(expression)
            cond.set_node(node)
            result.append(cond)
        elif node.type == NodeType.RETURN:
            # May return None
            prev = result[-1]
            if isinstance(prev, (OperationWithLValue)):
                r = Return(prev.lvalue)
                r.set_expression(expression)
                r.set_node(node)
                result.append(r)

    return result


###################################################################################
###################################################################################
# region Helpers
###################################################################################
###################################################################################


def is_value(ins: Operation) -> bool:
    if isinstance(ins, TmpCall):
        if isinstance(ins.ori, Member):
            if ins.ori.variable_right == "value":
                return True
    return False


def is_gas(ins: Operation) -> bool:
    if isinstance(ins, TmpCall):
        if isinstance(ins.ori, Member):
            if ins.ori.variable_right == "gas":
                return True
    return False


def _fits_under_integer(val: int, can_be_int: bool, can_be_uint: bool) -> list[str]:
    """
    Return the list of uint/int that can contain val

    :param val:
    :return:
    """
    ret: list[str] = []
    n = 8
    assert can_be_int | can_be_uint
    while n <= 256:
        if can_be_uint:
            if val <= 2**n - 1:
                ret.append(f"uint{n}")
        if can_be_int:
            if val <= (2**n) / 2 - 1:
                ret.append(f"int{n}")
        n = n + 8
    return ret


def _fits_under_byte(val: int | str) -> list[str]:
    """
    Return the list of byte that can contain val

    :param val:
    :return:
    """

    # If the value is written as an int, it can only be saved in one byte size
    # If its a string, it can be fitted under multiple values

    if isinstance(val, int):
        hex_val = hex(val)[2:]
        size = len(hex_val) // 2
        return [f"bytes{size}"]
    # val is a str
    length = len(val.encode("utf8"))
    return [f"bytes{f}" for f in range(length, 33)] + ["bytes"]


def _find_function_from_parameter(
    arguments: list[Variable], candidates: list[Function], full_comparison: bool
) -> Function | None:
    """
    Look for a function in candidates that can be the target based on the ir's call arguments

    Try the implicit type conversion for uint/int/bytes. Constant values can be both uint/int
    While variables stick to their base type, but can changed the size.
    If full_comparison is True it will do a comparison of all the arguments regardless if
    the candidate remained is one.

    :param arguments:
    :param candidates:
    :param full_comparison:
    :return:
    """
    type_args: list[str]
    for idx, arg in enumerate(arguments):
        if isinstance(arg, (list,)):
            type_args = [f"{get_type(arg[0].type)}[{len(arg)}]"]
        elif isinstance(arg, Function):
            type_args = [arg.signature_str]
        else:
            type_args = [get_type(arg.type)]

        arg_type = arg.type
        if isinstance(
            arg_type, ElementaryType
        ) and arg_type.type in ElementaryTypeInt + Uint + Byte + ["string"]:
            if isinstance(arg, Constant):
                value = arg.value
                can_be_uint = True
                can_be_int = True
            else:
                value = MaxValues[arg_type.type]
                can_be_uint = False
                can_be_int = False
                if arg_type.type in ElementaryTypeInt:
                    can_be_int = True
                elif arg_type.type in Uint:
                    can_be_uint = True

            if arg_type.type in ElementaryTypeInt + Uint:
                type_args = _fits_under_integer(value, can_be_int, can_be_uint)
            elif value is None and arg_type.type in ["bytes", "string"]:
                type_args = ["bytes", "string"]
            else:
                type_args = _fits_under_byte(value)
                if arg_type.type == "string":
                    type_args += ["string"]

        not_found = True
        candidates_kept: list[Function] = []
        for type_arg in type_args:
            if not not_found:
                break
            candidates_kept = []
            for candidate in candidates:
                param = get_type(candidate.parameters[idx].type)
                if param == type_arg:
                    not_found = False
                    candidates_kept.append(candidate)

            if len(candidates_kept) == 1 and not full_comparison:
                return candidates_kept[0]
        candidates = candidates_kept
    if len(candidates) == 1:
        return candidates[0]
    return None


def is_temporary(ins: Operation) -> bool:
    return isinstance(
        ins,
        (Argument, TmpNewElementaryType, TmpNewContract, TmpNewArray, TmpNewStructure),
    )


# endregion
###################################################################################
###################################################################################
# region Calls modification
###################################################################################
###################################################################################


def integrate_value_gas(result: list[Operation]) -> list[Operation]:
    """
    Integrate value and gas temporary arguments to call instruction
    """
    was_changed = True

    calls = []

    while was_changed:
        # We loop until we do not find any call to value or gas
        was_changed = False

        # Find all the assignments
        assigments = {}
        for i in result:
            if isinstance(i, OperationWithLValue) and i.lvalue:
                assigments[i.lvalue.name] = i
            if isinstance(i, TmpCall):
                if isinstance(i.called, Variable) and i.called.name in assigments:
                    ins_ori = assigments[i.called.name]
                    i.set_ori(ins_ori)

        to_remove = []
        variable_to_replace = {}

        # Replace call to value, gas to an argument of the real call
        for idx, ins in enumerate(result):
            # value can be shadowed, so we check that the prev ins
            # is an Argument
            if idx == 0:
                continue
            prev_ins = result[idx - 1]
            if is_value(ins) and isinstance(prev_ins, Argument):
                was_changed = True
                prev_ins.set_type(ArgumentType.VALUE)
                # Types checked by is_value
                prev_ins.call_id = ins.ori.variable_left.name  # type: ignore
                calls.append(ins.ori.variable_left)  # type: ignore
                to_remove.append(ins)
                variable_to_replace[ins.lvalue.name] = ins.ori.variable_left  # type: ignore
            elif is_gas(ins) and isinstance(prev_ins, Argument):
                was_changed = True
                prev_ins.set_type(ArgumentType.GAS)
                # Types checked by is_gas
                prev_ins.call_id = ins.ori.variable_left.name  # type: ignore
                calls.append(ins.ori.variable_left)  # type: ignore
                to_remove.append(ins)
                variable_to_replace[ins.lvalue.name] = ins.ori.variable_left  # type: ignore

        # Remove the call to value/gas instruction
        result = [i for i in result if i not in to_remove]

        # update the real call
        for ins in result:
            if isinstance(ins, TmpCall):
                # use of while if there redirections
                while ins.called.name in variable_to_replace:
                    was_changed = True
                    ins.call_id = variable_to_replace[ins.called.name].name
                    calls.append(ins.called)
                    ins.called = variable_to_replace[ins.called.name]
            if isinstance(ins, Argument):
                while ins.call_id in variable_to_replace:
                    was_changed = True
                    ins.call_id = variable_to_replace[ins.call_id].name

    calls = list({str(c) for c in calls})
    idx = 0
    calls_d = {}
    for call in calls:
        calls_d[str(call)] = idx
        idx = idx + 1

    return result


# endregion
###################################################################################
###################################################################################
# region Calls modification and Type propagation
###################################################################################
###################################################################################


def get_declared_param_names(
    ins: NewStructure
    | NewContract
    | InternalCall
    | LibraryCall
    | HighLevelCall
    | InternalDynamicCall
    | EventCall,
) -> list[list[str]] | None:
    """
    Given a call operation, return the list of parameter names, in the order
    listed in the function declaration.
    #### Parameters
    ins -
        The call instruction
    #### Possible Returns
    List[List[str]] -
        A list of list of parameters in declaration order. Note only if it's a Function there can be multiple list
    None -
        Workaround: Unable to obtain list of parameters in declaration order
    """
    if isinstance(ins, NewStructure):
        return [
            [x.name for x in ins.structure.elems_ordered if not isinstance(x.type, MappingType)]
        ]
    if isinstance(ins, (InternalCall, LibraryCall, HighLevelCall)):
        if isinstance(ins.function, Function):
            res = [[p.name for p in ins.function.parameters]]
            for f in ins.function.overrides:
                res.append([p.name for p in f.parameters])
            return res
        return None
    if isinstance(ins, InternalDynamicCall):
        return [[p.name for p in ins.function_type.params]]

    assert isinstance(ins, (EventCall, NewContract))
    return None


def reorder_arguments(
    args: list[Variable], call_names: list[str], decl_names: list[list[str]]
) -> list[Variable]:
    """
    Reorder named struct constructor arguments so that they match struct declaration ordering rather
    than call ordering
    E.g. for `struct S { int x; int y; }` we reorder `S({y : 2, x : 3})` to `S(3, 2)`
    #### Parameters
    args -
        Arguments to constructor call, in call order
    names -
        Parameter names in call order
    decl_names -
        List of list of parameter names in declaration order
    #### Returns
    Reordered arguments to constructor call, now in declaration order
    """
    assert len(args) == len(call_names)
    for names in decl_names:
        assert len(call_names) == len(names)

    args_ret = []
    index_seen = []  # Will have the correct index order

    # We search for declaration of parameters that is fully compatible with the call arguments
    # that is why if a arg name is not present in the call name we break and clear index_seen
    # Known issue if the overridden function reuse the same parameters' name but in different positions
    for names in decl_names:
        if len(index_seen) == len(args):
            break

        for n in names:
            try:
                ind = call_names.index(n)
                if ind in index_seen:
                    continue
            except ValueError:
                index_seen.clear()
                break
            index_seen.append(ind)

    for ind in index_seen:
        args_ret.append(args[ind])

    return args_ret


def propagate_type_and_convert_call(result: list[Operation], node: "Node") -> list[Operation]:
    """
    Propagate the types variables and convert tmp call to real call operation
    """
    calls_value = {}
    calls_gas = {}

    call_data = []

    idx = 0
    # use of while len() as result can be modified during the iteration
    while idx < len(result):
        ins = result[idx]

        if isinstance(ins, TmpCall):
            # If the node.function is a FunctionTopLevel, then it does not have a contract
            contract = (
                node.function.contract if isinstance(node.function, FunctionContract) else None
            )
            new_ins = extract_tmp_call(ins, contract)
            if new_ins:
                new_ins.set_node(ins.node)
                ins = new_ins
                result[idx] = ins

        # If there's two consecutive type conversions, keep only the last
        # ARG_CALL x                             call_data = [x]
        # TMP_4 = CONVERT x to Fix               call_data = []
        # TMP_5 = CONVERT TMP_4 to uint192       call_data = [TMP_5]
        if isinstance(ins, TypeConversion) and ins.variable in call_data:
            call_data.remove(ins.variable)

        if isinstance(ins, Argument):
            # In case of duplicate arguments we overwrite the value
            # This can happen because of addr.call.value(1).value(2)
            if ins.get_type() in [ArgumentType.GAS]:
                calls_gas[ins.call_id] = ins.argument
            elif ins.get_type() in [ArgumentType.VALUE]:
                calls_value[ins.call_id] = ins.argument
            else:
                assert ins.get_type() == ArgumentType.CALL
                call_data.append(ins.argument)

        if isinstance(ins, (HighLevelCall, NewContract, InternalDynamicCall)):
            if ins.call_id in calls_value:
                ins.call_value = calls_value[ins.call_id]
            if ins.call_id in calls_gas and isinstance(ins, (HighLevelCall, InternalDynamicCall)):
                ins.call_gas = calls_gas[ins.call_id]

        if isinstance(ins, Call) and (ins.names is not None):
            assert isinstance(
                ins,
                (
                    NewStructure,
                    NewContract,
                    InternalCall,
                    LibraryCall,
                    HighLevelCall,
                    InternalDynamicCall,
                    EventCall,
                ),
            )
            decl_param_names = get_declared_param_names(ins)
            if decl_param_names is not None:
                call_data = reorder_arguments(call_data, ins.names, decl_param_names)

        if isinstance(ins, (Call, NewContract, NewStructure)):
            # We might have stored some arguments for libraries
            if ins.arguments:
                call_data = ins.arguments + call_data
            ins.arguments = call_data
            call_data = []

        if is_temporary(ins):
            del result[idx]
            continue

        new_ins = propagate_types(ins, node)
        if new_ins:
            if isinstance(new_ins, (list,)):
                for new_ins_ in new_ins:
                    new_ins_.set_node(ins.node)
                del result[idx]
                for i, ins in enumerate(new_ins):
                    result.insert(idx + i, ins)
                idx = idx + len(new_ins) - 1
            else:
                new_ins.set_node(ins.node)
                result[idx] = new_ins
        idx = idx + 1
    return result


def _convert_type_contract(ir: Member) -> Assignment:
    assert isinstance(ir.variable_left.type, TypeInformation)
    contract = ir.variable_left.type.type

    scope = ir.node.file_scope

    if ir.variable_right == "creationCode":
        bytecode = scope.bytecode_init(
            ir.node.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        assignment = Assignment(ir.lvalue, Constant(str(bytecode)), ElementaryType("bytes"))
        assignment.set_expression(ir.expression)
        assignment.set_node(ir.node)
        assignment.lvalue.set_type(ElementaryType("bytes"))
        return assignment
    if ir.variable_right == "runtimeCode":
        bytecode = scope.bytecode_runtime(
            ir.node.compilation_unit.crytic_compile_compilation_unit, contract.name
        )
        assignment = Assignment(ir.lvalue, Constant(str(bytecode)), ElementaryType("bytes"))
        assignment.set_expression(ir.expression)
        assignment.set_node(ir.node)
        assignment.lvalue.set_type(ElementaryType("bytes"))
        return assignment
    if ir.variable_right == "interfaceId":
        entry_points = contract.functions_entry_points
        interfaceId = 0
        for entry_point in entry_points:
            interfaceId = interfaceId ^ get_function_id(entry_point.full_name)
        assignment = Assignment(
            ir.lvalue,
            Constant(str(interfaceId), constant_type=ElementaryType("bytes4")),
            ElementaryType("bytes4"),
        )
        assignment.set_expression(ir.expression)
        assignment.set_node(ir.node)
        assignment.lvalue.set_type(ElementaryType("bytes4"))
        return assignment

    if ir.variable_right == "name":
        assignment = Assignment(ir.lvalue, Constant(contract.name), ElementaryType("string"))
        assignment.set_expression(ir.expression)
        assignment.set_node(ir.node)
        assignment.lvalue.set_type(ElementaryType("string"))
        return assignment

    raise SlithIRError(f"type({contract.name}).{ir.variable_right} is unknown")


def propagate_types(ir: Operation, node: "Node"):
    # propagate the type
    node_function = node.function

    using_for: USING_FOR = {}
    if isinstance(node_function, FunctionContract):
        using_for = node_function.contract.using_for_complete
    elif isinstance(node_function, FunctionTopLevel):
        using_for = node_function.using_for_complete

    if isinstance(ir, OperationWithLValue) and ir.lvalue:
        # Force assignment in case of missing previous correct type
        if not ir.lvalue.type:
            if isinstance(ir, Assignment):
                ir.lvalue.set_type(ir.rvalue.type)
            elif isinstance(ir, Binary):
                if BinaryType.return_bool(ir.type):
                    ir.lvalue.set_type(ElementaryType("bool"))
                else:
                    ir.lvalue.set_type(ir.variable_left.type)
            elif isinstance(ir, Delete):
                # nothing to propagate
                pass
            elif isinstance(ir, LibraryCall):
                return convert_type_library_call(ir, ir.destination)
            elif isinstance(ir, HighLevelCall):
                t = ir.destination.type
                # Temporary operation (they are removed later)
                if t is None:
                    return None

                if isinstance(t, ElementaryType) and t.name == "address":
                    if can_be_solidity_func(ir):
                        return convert_to_solidity_func(ir)

                # convert library or top level function
                if t in using_for or "*" in using_for:
                    new_ir = convert_to_library_or_top_level(ir, node, using_for)
                    if new_ir:
                        return new_ir

                # convert library function when used with "this"
                if (
                    isinstance(t, ElementaryType)
                    and t.name == "address"
                    and ir.destination.name == "this"
                    and UserDefinedType(node_function.contract) in using_for
                ):
                    new_ir = convert_to_library_or_top_level(ir, node, using_for)
                    if new_ir:
                        return new_ir

                if isinstance(t, UserDefinedType):
                    # UserdefinedType
                    t_type = t.type
                    if isinstance(t_type, Contract):
                        # the target contract of the IR is the t_type (the destination of the call)
                        return convert_type_of_high_and_internal_level_call(ir, t_type)

                # Convert HighLevelCall to LowLevelCall
                if (isinstance(t, ElementaryType) and t.name == "address") or (
                    isinstance(t, TypeAlias) and t.underlying_type.name == "address"
                ):
                    if ir.destination.name == "this":
                        # Cannot be a top level function with this.
                        assert isinstance(node_function, FunctionContract)
                        # the target contract is the contract itself
                        return convert_type_of_high_and_internal_level_call(
                            ir, node_function.contract
                        )
                    if can_be_low_level(ir):
                        return convert_to_low_level(ir)

                # Convert push operations
                # May need to insert a new operation
                # Which leads to return a list of operation
                if isinstance(t, ArrayType) or (
                    isinstance(t, ElementaryType) and t.type == "bytes"
                ):
                    # Solidity uses push
                    # Vyper uses append
                    if ir.function_name in ["push", "append"] and len(ir.arguments) <= 1:
                        return convert_to_push(ir, node)
                    if ir.function_name == "pop" and len(ir.arguments) == 0:
                        return convert_to_pop(ir, node)

            elif isinstance(ir, Index):
                if isinstance(ir.variable_left.type, MappingType):
                    ir.lvalue.set_type(ir.variable_left.type.type_to)
                elif isinstance(ir.variable_left.type, ArrayType):
                    ir.lvalue.set_type(ir.variable_left.type.type)

            elif isinstance(ir, InitArray):
                length = len(ir.init_values)
                t = ir.init_values[0].type
                ir.lvalue.set_type(ArrayType(t, length))
            elif isinstance(ir, InternalCall):
                # if its not a tuple, return a singleton
                if ir.function is None:
                    func = node.function
                    function_contract = (
                        func.contract if isinstance(func, FunctionContract) else None
                    )
                    # the target contract might be None if its a top level function
                    convert_type_of_high_and_internal_level_call(ir, function_contract)
                return_type = ir.function.return_type
                if return_type:
                    if len(return_type) == 1:
                        ir.lvalue.set_type(return_type[0])
                    elif len(return_type) > 1:
                        ir.lvalue.set_type(return_type)
                else:
                    ir.lvalue = None
            elif isinstance(ir, InternalDynamicCall):
                # if its not a tuple, return a singleton
                return_type = ir.function_type.return_type
                if return_type:
                    if len(return_type) == 1:
                        ir.lvalue.set_type(return_type[0])
                    else:
                        ir.lvalue.set_type(return_type)
                else:
                    ir.lvalue = None
            elif isinstance(ir, LowLevelCall):
                # Call are not yet converted
                # This should not happen
                assert False
            elif isinstance(ir, Member):
                # TODO we should convert the reference to a temporary if the member is a length or a balance
                if (
                    ir.variable_right == "length"
                    and not isinstance(ir.variable_left, Contract)
                    and isinstance(ir.variable_left.type, (ElementaryType, ArrayType))
                ):
                    new_length = Length(ir.variable_left, ir.lvalue)
                    assert ir.expression
                    new_length.set_expression(ir.expression)
                    new_length.lvalue.points_to = ir.variable_left
                    new_length.set_node(ir.node)
                    return new_length
                # This only happen for .balance/code/codehash access on a variable for which we dont know at
                # early parsing time the type
                # Like
                # function return_addr() internal returns(addresss)
                #
                # return_addr().balance
                # Here slithIR will incorrectly create a REF variable instead of a Temp variable
                # However this pattern does not appear so often
                if (
                    ir.variable_right.name in ["balance", "code", "codehash"]
                    and not isinstance(ir.variable_left, Contract)
                    and isinstance(ir.variable_left.type, ElementaryType)
                ):
                    name = ir.variable_right.name + "(address)"
                    sol_func = SolidityFunction(name)
                    s = SolidityCall(
                        sol_func,
                        1,
                        ir.lvalue,
                        sol_func.return_type,
                    )
                    s.arguments.append(ir.variable_left)
                    s.set_expression(ir.expression)
                    s.lvalue.set_type(sol_func.return_type)
                    s.set_node(ir.node)
                    return s
                if (
                    ir.variable_right == "codesize"
                    and not isinstance(ir.variable_left, Contract)
                    and isinstance(ir.variable_left.type, ElementaryType)
                ):
                    b = CodeSize(ir.variable_left, ir.lvalue)
                    b.set_expression(ir.expression)
                    b.set_node(ir.node)
                    return b
                if ir.variable_right == "selector" and isinstance(ir.variable_left, (CustomError)):
                    assignment = Assignment(
                        ir.lvalue,
                        Constant(
                            str(get_function_id(ir.variable_left.solidity_signature)),
                            ElementaryType("bytes4"),
                        ),
                        ElementaryType("bytes4"),
                    )
                    assignment.set_expression(ir.expression)
                    assignment.set_node(ir.node)
                    assignment.lvalue.set_type(ElementaryType("bytes4"))
                    return assignment

                if isinstance(ir.variable_right, (CustomError)):
                    assignment = Assignment(
                        ir.lvalue,
                        Constant(
                            str(get_function_id(ir.variable_left.solidity_signature)),
                            ElementaryType("bytes4"),
                        ),
                        ElementaryType("bytes4"),
                    )
                    assignment.set_expression(ir.expression)
                    assignment.set_node(ir.node)
                    assignment.lvalue.set_type(ElementaryType("bytes4"))
                    return assignment
                if ir.variable_right == "selector" and isinstance(ir.variable_left, (Event)):
                    # the event selector returns a bytes32, which is different from the error/function selector
                    # which returns a bytes4
                    assignment = Assignment(
                        ir.lvalue,
                        Constant(
                            str(get_event_id(ir.variable_left.full_name)), ElementaryType("bytes32")
                        ),
                        ElementaryType("bytes32"),
                    )
                    assignment.set_expression(ir.expression)
                    assignment.set_node(ir.node)
                    assignment.lvalue.set_type(ElementaryType("bytes32"))
                    return assignment
                if ir.variable_right == "selector" and (
                    isinstance(ir.variable_left.type, (Function))
                ):
                    assignment = Assignment(
                        ir.lvalue,
                        Constant(
                            str(get_function_id(ir.variable_left.type.full_name)),
                            ElementaryType("bytes4"),
                        ),
                        ElementaryType("bytes4"),
                    )
                    assignment.set_expression(ir.expression)
                    assignment.set_node(ir.node)
                    assignment.lvalue.set_type(ElementaryType("bytes4"))
                    return assignment
                if isinstance(ir.variable_left, TemporaryVariable) and isinstance(
                    ir.variable_left.type, TypeInformation
                ):
                    return _convert_type_contract(ir)
                left = ir.variable_left
                t = None
                ir_func = ir.node.function
                # Handling of this.function_name usage
                if (
                    left == SolidityVariable("this")
                    and isinstance(ir.variable_right, Constant)
                    and isinstance(ir_func, FunctionContract)
                    and str(ir.variable_right) in [x.name for x in ir_func.contract.functions]
                ):
                    # Assumption that this.function_name can only compile if
                    # And the contract does not have two functions starting with function_name
                    # Otherwise solc raises:
                    # Error: Member "f" not unique after argument-dependent lookup in contract
                    targeted_function = next(
                        x for x in ir_func.contract.functions if x.name == str(ir.variable_right)
                    )
                    ir.lvalue.set_type(targeted_function)
                elif isinstance(left, (Variable, SolidityVariable)):
                    t = left.type
                elif isinstance(left, (Contract, Enum, Structure)):
                    t = UserDefinedType(left)
                # can be None due to temporary operation
                if t:
                    if isinstance(t, UserDefinedType):
                        # UserdefinedType
                        type_t = t.type
                        if isinstance(type_t, Enum):
                            ir.lvalue.set_type(t)
                        elif isinstance(type_t, Structure):
                            elems = type_t.elems
                            for elem in elems:
                                if elem == ir.variable_right:
                                    ir.lvalue.set_type(elems[elem].type)
                        else:
                            assert isinstance(type_t, Contract)
                            # Allow type propagation as a Function
                            # Only for reference variables
                            # This allows to track the selector keyword
                            # We don't need to check for function collision, as solc prevents the use of selector
                            # if there are multiple functions with the same name
                            f = next(
                                (f for f in type_t.functions if f.name == ir.variable_right),
                                None,
                            )
                            if f:
                                ir.lvalue.set_type(f)
                            else:
                                # Allow propagation for variable access through contract's name
                                # like Base_contract.my_variable
                                v = next(
                                    (
                                        v
                                        for v in type_t.state_variables
                                        if v.name == ir.variable_right
                                    ),
                                    None,
                                )
                                if v:
                                    ir.lvalue.set_type(v.type)
            elif isinstance(ir, NewArray):
                ir.lvalue.set_type(ir.array_type)
            elif isinstance(ir, NewContract):
                ir.lvalue.set_type(ir.contract_name)
            elif isinstance(ir, NewElementaryType):
                ir.lvalue.set_type(ir.type)
            elif isinstance(ir, NewStructure):
                ir.lvalue.set_type(UserDefinedType(ir.structure))
            elif isinstance(ir, Send):
                ir.lvalue.set_type(ElementaryType("bool"))
            elif isinstance(ir, SolidityCall):
                if ir.function.name in ["type(address)", "type()"]:
                    ir.function.return_type = [TypeInformation(ir.arguments[0])]
                return_type = ir.function.return_type
                if len(return_type) == 1:
                    ir.lvalue.set_type(return_type[0])
                elif len(return_type) > 1:
                    ir.lvalue.set_type(return_type)
            elif isinstance(ir, TypeConversion):
                ir.lvalue.set_type(ir.type)
            elif isinstance(ir, Unary):
                ir.lvalue.set_type(ir.rvalue.type)
            elif isinstance(ir, Unpack):
                types = ir.tuple.type.type
                idx = ir.index
                t = types[idx]
                ir.lvalue.set_type(t)
            elif isinstance(
                ir,
                (
                    Argument,
                    TmpCall,
                    TmpNewArray,
                    TmpNewContract,
                    TmpNewStructure,
                    TmpNewElementaryType,
                ),
            ):
                # temporary operation; they will be removed
                pass
            else:
                raise SlithIRError(f"Not handling {type(ir)} during type propagation")
    return None


def extract_tmp_call(ins: TmpCall, contract: Contract | None) -> Call | Nop:
    assert isinstance(ins, TmpCall)
    if isinstance(ins.called, Variable) and isinstance(ins.called.type, FunctionType):
        # If the call is made to a variable member, where the member is this
        # We need to convert it to a HighLelelCall and not an internal dynamic call
        if isinstance(ins.ori, Member) and ins.ori.variable_left == SolidityVariable("this"):
            pass
        else:
            call = InternalDynamicCall(ins.lvalue, ins.called, ins.called.type)
            call.set_expression(ins.expression)
            call.call_id = ins.call_id
            return call
    if isinstance(ins.ori, Member):
        # If there is a call on an inherited contract, it is an internal call or an event
        if contract and ins.ori.variable_left in contract.inheritance + [contract]:
            if str(ins.ori.variable_right) in [f.name for f in contract.functions]:
                internalcall = InternalCall(
                    (ins.ori.variable_right, ins.ori.variable_left.name),
                    ins.nbr_arguments,
                    ins.lvalue,
                    ins.type_call,
                )
                internalcall.set_expression(ins.expression)
                internalcall.call_id = ins.call_id
                return internalcall
            if str(ins.ori.variable_right) in [f.name for f in contract.events]:
                eventcall = EventCall(ins.ori.variable_right)
                eventcall.set_expression(ins.expression)
                eventcall.call_id = ins.call_id
                return eventcall
        if isinstance(ins.ori.variable_left, Contract):
            st = ins.ori.variable_left.get_structure_from_name(ins.ori.variable_right)
            if st:
                op = NewStructure(st, ins.lvalue, names=ins.names)
                op.set_expression(ins.expression)
                op.call_id = ins.call_id
                return op

            # For event emit through library
            # lib L { event E()}
            # ...
            # emit L.E();
            if str(ins.ori.variable_right) in [f.name for f in ins.ori.variable_left.events]:
                eventcall = EventCall(ins.ori.variable_right)
                eventcall.set_expression(ins.expression)
                eventcall.call_id = ins.call_id
                return eventcall

            # lib Lib { error Error()} ... revert Lib.Error()
            if str(ins.ori.variable_right) in ins.ori.variable_left.custom_errors_as_dict:
                custom_error = ins.ori.variable_left.custom_errors_as_dict[
                    str(ins.ori.variable_right)
                ]
                assert isinstance(
                    custom_error,
                    CustomError,
                )
                sol_function = SolidityCustomRevert(custom_error)
                solidity_call = SolidityCall(
                    sol_function, ins.nbr_arguments, ins.lvalue, ins.type_call
                )
                solidity_call.set_expression(ins.expression)
                return solidity_call

            libcall = LibraryCall(
                ins.ori.variable_left,
                ins.ori.variable_right,
                ins.nbr_arguments,
                ins.lvalue,
                ins.type_call,
                names=ins.names,
            )
            libcall.set_expression(ins.expression)
            libcall.call_id = ins.call_id
            return libcall
        if isinstance(ins.ori.variable_left, Function):
            # Support for library call where the parameter is a function
            # We could merge this with the standard library handling
            # Except that we will have some troubles with using_for
            # As the type of the function will not match function()
            # Additionally we do not have a correct view on the parameters of the tmpcall
            # At this level
            #
            # library FunctionExtensions {
            #     function h(function() internal _t, uint8) internal {  }
            # }
            # contract FunctionMembers {
            #     using FunctionExtensions for function();
            #
            #     function f() public {
            #         f.h(1);
            #     }
            # }
            node_func = ins.node.function
            using_for = (
                node_func.contract.using_for_complete
                if isinstance(node_func, FunctionContract)
                else {}
            )

            targeted_libraries = (
                [] + using_for.get("*", []) + using_for.get(FunctionType([], []), [])
            )
            lib_contract: Contract
            candidates = []
            for lib_contract_type in targeted_libraries:
                if not isinstance(lib_contract_type, UserDefinedType) and isinstance(
                    lib_contract_type.type, Contract
                ):
                    continue
                if isinstance(lib_contract_type, FunctionContract):
                    # Using for with list of functions, this is the function called
                    candidates.append(lib_contract_type)
                else:
                    lib_contract = lib_contract_type.type
                    for lib_func in lib_contract.functions:
                        if lib_func.name == ins.ori.variable_right:
                            candidates.append(lib_func)

            if len(candidates) == 1:
                lib_func = candidates[0]
                # Library must be from a contract
                assert isinstance(lib_func, FunctionContract)
                lib_call = LibraryCall(
                    lib_func.contract,
                    Constant(lib_func.name),
                    len(lib_func.parameters),
                    ins.lvalue,
                    "d",
                    names=ins.names,
                )
                lib_call.set_expression(ins.expression)
                lib_call.set_node(ins.node)
                lib_call.call_gas = ins.call_gas
                lib_call.call_id = ins.call_id
                lib_call.set_node(ins.node)
                lib_call.function = lib_func
                lib_call.arguments.append(ins.ori.variable_left)
                return lib_call
            # We do not support something lik
            # library FunctionExtensions {
            #     function h(function() internal _t, uint8) internal {  }
            #     function h(function() internal _t, bool) internal {  }
            # }
            # contract FunctionMembers {
            #     using FunctionExtensions for function();
            #
            #     function f() public {
            #         f.h(1);
            #     }
            # }
            to_log = "Slither does not support dynamic functions to libraries if functions have the same name"
            to_log += f"{[candidate.full_name for candidate in candidates]}"
            raise SlithIRError(to_log)
        if isinstance(ins.ori.variable_left, SolidityImportPlaceHolder):
            # For top level import, where the import statement renames the filename
            # See https://blog.soliditylang.org/2020/09/02/solidity-0.7.1-release-announcement/

            current_path = Path(ins.ori.variable_left.source_mapping.filename.absolute).parent
            target = str(
                Path(current_path, ins.ori.variable_left.import_directive.filename).absolute()
            )
            top_level_function_targets = [
                f
                for f in ins.compilation_unit.functions_top_level
                if f.source_mapping.filename.absolute == target
                and f.name == ins.ori.variable_right
                and len(f.parameters) == ins.nbr_arguments
            ]

            internalcall = InternalCall(
                (ins.ori.variable_right, ins.ori.variable_left.name),
                ins.nbr_arguments,
                ins.lvalue,
                ins.type_call,
            )
            internalcall.set_expression(ins.expression)
            internalcall.call_id = ins.call_id
            internalcall.function_candidates = top_level_function_targets
            return internalcall

        if ins.ori.variable_left == ElementaryType("bytes") and ins.ori.variable_right == Constant(
            "concat"
        ):
            s = SolidityCall(
                SolidityFunction("bytes.concat()"),
                ins.nbr_arguments,
                ins.lvalue,
                ins.type_call,
            )
            s.set_expression(ins.expression)
            return s

        if ins.ori.variable_left == ElementaryType("string") and ins.ori.variable_right == Constant(
            "concat"
        ):
            s = SolidityCall(
                SolidityFunction("string.concat()"),
                ins.nbr_arguments,
                ins.lvalue,
                ins.type_call,
            )
            s.set_expression(ins.expression)
            return s

        msgcall = HighLevelCall(
            ins.ori.variable_left,
            ins.ori.variable_right,
            ins.nbr_arguments,
            ins.lvalue,
            ins.type_call,
            names=ins.names,
        )
        msgcall.call_id = ins.call_id

        if ins.call_gas:
            msgcall.call_gas = ins.call_gas
        if ins.call_value:
            msgcall.call_value = ins.call_value
        msgcall.set_expression(ins.expression)

        return msgcall

    if isinstance(ins.ori, TmpCall):
        r = extract_tmp_call(ins.ori, contract)
        r.set_node(ins.node)
        return r
    if isinstance(ins.called, SolidityVariableComposed):
        if str(ins.called) == "block.blockhash":
            ins.called = SolidityFunction("blockhash(uint256)")

    if isinstance(ins.called, SolidityFunction):
        s = SolidityCall(ins.called, ins.nbr_arguments, ins.lvalue, ins.type_call)
        s.set_expression(ins.expression)
        return s

    if isinstance(ins.called, CustomError):
        sol_function = SolidityCustomRevert(ins.called)
        s = SolidityCall(sol_function, ins.nbr_arguments, ins.lvalue, ins.type_call)
        s.set_expression(ins.expression)
        return s

    if isinstance(ins.ori, TmpNewElementaryType):
        n = NewElementaryType(ins.ori.type, ins.lvalue)
        n.set_expression(ins.expression)
        return n

    if isinstance(ins.ori, TmpNewContract):
        op = NewContract(ins.ori.contract_name, ins.lvalue)
        op.set_expression(ins.expression)
        op.call_id = ins.call_id
        if ins.call_value:
            op.call_value = ins.call_value
        if ins.call_salt:
            op.call_salt = ins.call_salt
        return op

    if isinstance(ins.ori, TmpNewArray):
        n = NewArray(ins.ori.array_type, ins.lvalue)
        n.set_expression(ins.expression)
        return n

    if isinstance(ins.called, Structure):
        op = NewStructure(ins.called, ins.lvalue, names=ins.names)
        op.set_expression(ins.expression)
        op.call_id = ins.call_id
        op.set_expression(ins.expression)
        return op

    if isinstance(ins.called, Event):
        e = EventCall(ins.called.name)
        e.set_expression(ins.expression)
        return e

    if isinstance(ins.called, Contract):
        # Called a base constructor, where there is no constructor
        if ins.called.constructor is None:
            return Nop()
        # Case where:
        # contract A{ constructor(uint) }
        # contract B is A {}
        # contract C is B{ constructor() A(10) B() {}
        # C calls B(), which does not exist
        # Ideally we should compare here for the parameters types too
        if len(ins.called.constructor.parameters) != ins.nbr_arguments:
            return Nop()
        internalcall = InternalCall(
            ins.called.constructor, ins.nbr_arguments, ins.lvalue, ins.type_call, ins.names
        )
        internalcall.call_id = ins.call_id
        internalcall.set_expression(ins.expression)
        return internalcall

    raise SlithIRError(f"Not extracted {type(ins.called)} {ins}")


# endregion
###################################################################################
###################################################################################
# region Conversion operations
###################################################################################
###################################################################################


def can_be_low_level(ir: HighLevelCall) -> bool:
    return ir.function_name in [
        "transfer",
        "send",
        "call",
        "delegatecall",
        "callcode",
        "staticcall",
        "raw_call",
    ]


def convert_to_low_level(
    ir: HighLevelCall,
) -> Send | LowLevelCall | Transfer:
    """
    Convert to a transfer/send/or low level call
    The function assume to receive a correct IR
    The checks must be done by the caller

    Must be called after can_be_low_level
    """
    if ir.function_name == "transfer":
        assert len(ir.arguments) == 1
        prev_ir = ir
        ir = Transfer(ir.destination, ir.arguments[0])
        ir.set_expression(prev_ir.expression)
        ir.set_node(prev_ir.node)
        return ir
    if ir.function_name == "send":
        assert len(ir.arguments) == 1
        prev_ir = ir
        ir = Send(ir.destination, ir.arguments[0], ir.lvalue)
        ir.set_expression(prev_ir.expression)
        ir.set_node(prev_ir.node)
        ir.lvalue.set_type(ElementaryType("bool"))
        return ir
    if ir.function_name in ["call", "delegatecall", "callcode", "staticcall", "raw_call"]:
        new_ir = LowLevelCall(
            ir.destination, ir.function_name, ir.nbr_arguments, ir.lvalue, ir.type_call
        )
        new_ir.call_gas = ir.call_gas
        new_ir.call_value = ir.call_value
        new_ir.arguments = ir.arguments
        # TODO fix this for Vyper
        if ir.node.compilation_unit.solc_version >= "0.5":
            new_ir.lvalue.set_type([ElementaryType("bool"), ElementaryType("bytes")])
        else:
            new_ir.lvalue.set_type(ElementaryType("bool"))
        new_ir.set_expression(ir.expression)
        new_ir.set_node(ir.node)
        return new_ir
    raise SlithIRError(f"Incorrect conversion to low level {ir}")


def can_be_solidity_func(ir: HighLevelCall) -> bool:
    if not isinstance(ir, HighLevelCall):
        return False
    return ir.destination.name == "abi" and ir.function_name in [
        "encode",
        "encodePacked",
        "encodeWithSelector",
        "encodeWithSignature",
        "encodeCall",
        "decode",
    ]


def convert_to_solidity_func(
    ir: HighLevelCall,
) -> SolidityCall:
    """
    Must be called after can_be_solidity_func
    :param ir:
    :return:
    """
    call = SolidityFunction(f"abi.{ir.function_name}()")
    new_ir = SolidityCall(call, ir.nbr_arguments, ir.lvalue, ir.type_call)
    new_ir.arguments = ir.arguments
    new_ir.set_expression(ir.expression)
    new_ir.set_node(ir.node)
    if isinstance(call.return_type, list) and len(call.return_type) == 1:
        new_ir.lvalue.set_type(call.return_type[0])
    elif (
        isinstance(new_ir.lvalue, TupleVariable)
        and call == SolidityFunction("abi.decode()")
        and len(new_ir.arguments) == 2
        and isinstance(new_ir.arguments[1], list)
    ):
        types = []
        for arg_type in new_ir.arguments[1]:
            decode_type = arg_type
            if isinstance(decode_type, (Structure, Enum, Contract)):
                decode_type = UserDefinedType(decode_type)
            types.append(decode_type)
        new_ir.lvalue.set_type(types)
    # abi.decode where the type to decode is a singleton
    # abi.decode(a, (uint))
    elif call == SolidityFunction("abi.decode()") and len(new_ir.arguments) == 2:
        # If the variable is a referenceVariable, we are lost
        # See https://github.com/crytic/slither/issues/566 for potential solutions
        if not isinstance(new_ir.arguments[1], ReferenceVariable):
            decode_type = new_ir.arguments[1]
            if isinstance(decode_type, (Structure, Enum, Contract)):
                decode_type = UserDefinedType(decode_type)
            new_ir.lvalue.set_type(decode_type)
    else:
        new_ir.lvalue.set_type(call.return_type)
    return new_ir


def convert_to_push_expand_arr(
    ir: HighLevelCall, node: "Node", ret: list[Any]
) -> TemporaryVariable:
    arr = ir.destination

    length = ReferenceVariable(node)
    length.set_type(ElementaryType("uint256"))

    ir_length = Length(arr, length)
    ir_length.set_expression(ir.expression)
    ir_length.set_node(ir.node)
    ir_length.lvalue.points_to = arr
    ret.append(ir_length)

    length_val = TemporaryVariable(node)
    length_val.set_type(ElementaryType("uint256"))
    ir_get_length = Assignment(length_val, length, ElementaryType("uint256"))
    ir_get_length.set_expression(ir.expression)
    ir_get_length.set_node(ir.node)
    ret.append(ir_get_length)

    new_length_val = TemporaryVariable(node)
    ir_add_1 = Binary(
        new_length_val, length_val, Constant("1", ElementaryType("uint256")), BinaryType.ADDITION
    )
    ir_add_1.set_expression(ir.expression)
    ir_add_1.set_node(ir.node)
    ret.append(ir_add_1)

    ir_assign_length = Assignment(length, new_length_val, ElementaryType("uint256"))
    ir_assign_length.set_expression(ir.expression)
    ir_assign_length.set_node(ir.node)
    ret.append(ir_assign_length)

    return length_val


def convert_to_push_set_val(
    ir: HighLevelCall,
    node: "Node",
    length_val: TemporaryVariable,
    ret: list[Length | Assignment | Binary],
) -> None:
    arr = ir.destination

    new_type = ir.destination.type.type

    element_to_add = ReferenceVariable(node)
    element_to_add.set_type(new_type)
    ir_assign_element_to_add = Index(element_to_add, arr, length_val)
    ir_assign_element_to_add.set_expression(ir.expression)
    ir_assign_element_to_add.set_node(ir.node)
    ret.append(ir_assign_element_to_add)

    if len(ir.arguments) > 0:
        assign_value = ir.arguments[0]
        if isinstance(assign_value, list):
            assign_value = TemporaryVariable(node)
            assign_value.set_type(element_to_add.type)
            ir_assign_value = InitArray(ir.arguments[0], assign_value)
            ir_assign_value.set_expression(ir.expression)
            ir_assign_value.set_node(ir.node)
            ret.append(ir_assign_value)

        ir_assign_value = Assignment(element_to_add, assign_value, assign_value.type)
        ir_assign_value.set_expression(ir.expression)
        ir_assign_value.set_node(ir.node)
        ret.append(ir_assign_value)
    else:
        new_element = ir.lvalue
        new_element.set_type(new_type)
        ir_assign_value = Assignment(new_element, element_to_add, new_type)
        ir_assign_value.set_expression(ir.expression)
        ir_assign_value.set_node(ir.node)
        ret.append(ir_assign_value)


def convert_to_push(
    ir: HighLevelCall, node: "Node"
) -> list[Length | Assignment | Binary | Index | InitArray]:
    """
    Convert a call to a series of operations to push a new value onto the array

    The function assume to receive a correct IR
    The checks must be done by the caller

    May necessitate to create an intermediate operation (InitArray)
    Necessitate to return the length (see push documentation)
    As a result, the function return may return a list
    """

    ret = []

    length_val = convert_to_push_expand_arr(ir, node, ret)
    convert_to_push_set_val(ir, node, length_val, ret)

    return ret


def convert_to_pop(ir: HighLevelCall, node: "Node") -> list[Operation]:
    """
    Convert pop operators
    Return a list of 6 operations
    """

    ret: list[Operation] = []

    arr = ir.destination
    length = ReferenceVariable(node)
    length.set_type(ElementaryType("uint256"))

    ir_length = Length(arr, length)
    assert ir.expression
    ir_length.set_expression(ir.expression)
    ir_length.set_node(ir.node)
    length.points_to = arr
    ret.append(ir_length)

    val = TemporaryVariable(node)

    ir_sub_1 = Binary(val, length, Constant("1", ElementaryType("uint256")), BinaryType.SUBTRACTION)
    ir_sub_1.set_expression(ir.expression)
    ir_sub_1.set_node(ir.node)
    ret.append(ir_sub_1)

    element_to_delete = ReferenceVariable(node)
    ir_assign_element_to_delete = Index(element_to_delete, arr, val)
    # TODO the following is equivalent to length.points_to = arr
    # Should it be removed?
    ir_length.lvalue.points_to = arr
    # Note bytes is an ElementaryType not ArrayType and bytes1 should be returned
    # since bytes is bytes1[] without padding between the elements
    # while in other cases such as uint256[] (ArrayType) we use ir.destination.type.type
    # in this way we will have the type always set to the corresponding ElementaryType
    element_to_delete.set_type(
        ElementaryType("bytes1")
        if isinstance(ir.destination.type, ElementaryType)
        else ir.destination.type.type
    )
    ir_assign_element_to_delete.set_expression(ir.expression)
    ir_assign_element_to_delete.set_node(ir.node)
    ret.append(ir_assign_element_to_delete)

    ir_delete = Delete(element_to_delete, element_to_delete)
    ir_delete.set_expression(ir.expression)
    ir_delete.set_node(ir.node)
    ret.append(ir_delete)

    length_to_assign = ReferenceVariable(node)
    length_to_assign.set_type(ElementaryType("uint256"))
    ir_length = Length(arr, length_to_assign)
    ir_length.set_expression(ir.expression)
    length_to_assign.points_to = arr
    ir_length.set_node(ir.node)
    ret.append(ir_length)

    ir_assign_length = Assignment(length_to_assign, val, ElementaryType("uint256"))
    ir_assign_length.set_expression(ir.expression)
    ir_assign_length.set_node(ir.node)
    ret.append(ir_assign_length)

    return ret


def look_for_library_or_top_level(
    ir: HighLevelCall,
    using_for,
    t: UserDefinedType | ElementaryType | str | TypeAliasTopLevel,
) -> LibraryCall | InternalCall | None:
    for destination in using_for[t]:
        if isinstance(destination, FunctionTopLevel) and destination.name == ir.function_name:
            arguments = [ir.destination] + ir.arguments
            if (
                len(destination.parameters) == len(arguments)
                and _find_function_from_parameter(arguments, [destination], True) is not None
            ):
                internalcall = InternalCall(destination, ir.nbr_arguments, ir.lvalue, ir.type_call)
                internalcall.set_expression(ir.expression)
                internalcall.set_node(ir.node)
                internalcall.arguments = [ir.destination] + ir.arguments
                return_type = internalcall.function.return_type
                if return_type:
                    if len(return_type) == 1:
                        internalcall.lvalue.set_type(return_type[0])
                    elif len(return_type) > 1:
                        internalcall.lvalue.set_type(return_type)
                else:
                    internalcall.lvalue = None
                return internalcall

        lib_contract = None
        if isinstance(destination, FunctionContract) and destination.contract.is_library:
            lib_contract = destination.contract
        elif isinstance(destination, UserDefinedType) and isinstance(destination.type, Contract):
            lib_contract = destination.type

        if lib_contract:
            lib_call = LibraryCall(
                lib_contract,
                ir.function_name,
                ir.nbr_arguments,
                ir.lvalue,
                ir.type_call,
                names=ir.names,
            )
            lib_call.set_expression(ir.expression)
            lib_call.set_node(ir.node)
            lib_call.call_gas = ir.call_gas
            lib_call.arguments = [ir.destination] + ir.arguments
            new_ir = convert_type_library_call(lib_call, lib_contract)
            if new_ir:
                new_ir.set_node(ir.node)
                return new_ir
    return None


def convert_to_library_or_top_level(
    ir: HighLevelCall, node: "Node", using_for
) -> LibraryCall | InternalCall | None:
    t = ir.destination.type
    if t in using_for:
        new_ir = look_for_library_or_top_level(ir, using_for, t)
        if new_ir:
            return new_ir

    if "*" in using_for:
        new_ir = look_for_library_or_top_level(ir, using_for, "*")
        if new_ir:
            return new_ir

    if (
        isinstance(t, ElementaryType)
        and t.name == "address"
        and ir.destination.name == "this"
        and UserDefinedType(node.function.contract) in using_for
    ):
        new_ir = look_for_library_or_top_level(
            ir, using_for, UserDefinedType(node.function.contract)
        )
        if new_ir:
            return new_ir

    return None


def get_type(
    t: UserDefinedType | ElementaryType,
) -> str:
    """
    Convert a type to a str
    If the instance is a Contract, return 'address' instead
    """
    if isinstance(t, UserDefinedType):
        if isinstance(t.type, Contract):
            return "address"
    return str(t)


def _can_be_implicitly_converted(source: str, target: str) -> bool:
    if source in ElementaryTypeInt and target in ElementaryTypeInt:
        return int(source[3:]) <= int(target[3:])
    if source in Uint and target in Uint:
        return int(source[4:]) <= int(target[4:])
    return source == target


def convert_type_library_call(ir: HighLevelCall, lib_contract: Contract) -> LibraryCall | None:
    func = None
    candidates = [
        f
        for f in lib_contract.functions
        if f.name == ir.function_name
        and not f.is_shadowed
        and len(f.parameters) == len(ir.arguments)
    ]

    if len(candidates) == 1:
        func = candidates[0]
    # We can discard if there are arguments here because libraries only support constant variables
    # And constant variables cannot have non-value type
    # i.e. "uint[2] constant arr = [1,2];" is not possible in Solidity
    # If this were to change, the following condition might be broken
    if func is None and not ir.arguments:
        # TODO: handle collision with multiple state variables/functions
        func = lib_contract.get_state_variable_from_name(ir.function_name)
    if func is None and candidates:
        func = _find_function_from_parameter(ir.arguments, candidates, False)

    # In case of multiple binding to the same type
    # TODO: this part might not be needed with _find_function_from_parameter
    if not func:
        # specific lookup when the compiler does implicit conversion
        # for example
        # myFunc(uint)
        # can be called with an uint8
        for function in lib_contract.functions:
            if function.name == ir.function_name and len(function.parameters) == len(ir.arguments):
                func = function
                break
    if not func:
        return None
    ir.function = func
    if isinstance(func, Function):
        t = func.return_type
        # if its not a tuple, return a singleton
        if t and len(t) == 1:
            t = t[0]
    else:
        # otherwise its a variable (getter)
        t = func.type
    if t:
        ir.lvalue.set_type(t)
    else:
        ir.lvalue = None
    return ir


def _convert_to_structure_to_list(return_type: Type) -> list[Type]:
    """
    Convert structure elements types to a list of types
    Recursive function

    :param return_type:
    :return:
    """
    if isinstance(return_type, UserDefinedType) and isinstance(return_type.type, Structure):
        ret = []
        for v in return_type.type.elems_ordered:
            ret += _convert_to_structure_to_list(v.type)
        return ret
    # Mapping and arrays are not included in external call
    #
    # contract A{
    #
    #     struct St{
    #         uint a;
    #         uint b;
    #         mapping(uint => uint) map;
    #         uint[] array;
    #     }
    #
    #     mapping (uint => St) public st;
    #
    # }
    #
    # contract B{
    #
    #     function f(A a) public{
    #         (uint a, uint b) = a.st(0);
    #     }
    # }
    if isinstance(return_type, (MappingType, ArrayType)):
        return []

    assert isinstance(return_type, (ElementaryType, UserDefinedType, TypeAlias))
    return [return_type]


def convert_type_of_high_and_internal_level_call(
    ir: Operation, contract: Contract | None
) -> Operation | None:
    """
    Convert the IR type based on heuristic

    Args:
        ir: target
        contract: optional contract. This should be the target of the IR. It will be used to look up potential functions

    Returns:
        Potential new IR
    """

    func = None
    if isinstance(ir, InternalCall):
        candidates: list[Function]
        if ir.function_candidates:
            # This path is taken only for SolidityImportPlaceHolder
            # Here we have already done a filtering on the potential targets
            candidates = ir.function_candidates
        else:
            candidates = [
                f
                for f in contract.functions
                if f.name == ir.function_name
                and f.contract_declarer.name == ir.contract_name
                and len(f.parameters) == len(ir.arguments)
            ]

            for import_statement in contract.file_scope.imports:
                if (
                    import_statement.alias is not None
                    and import_statement.alias == ir.contract_name
                ):
                    imported_scope = contract.compilation_unit.get_scope(import_statement.filename)
                    candidates += [
                        f
                        for f in list(imported_scope.functions)
                        if f.name == ir.function_name and len(f.parameters) == len(ir.arguments)
                    ]

        func = _find_function_from_parameter(ir.arguments, candidates, False)

        if not func:
            assert contract
            func = contract.get_state_variable_from_name(ir.function_name)
    else:
        assert isinstance(ir, HighLevelCall)
        assert contract
        candidates = [
            f
            for f in contract.functions
            if f.name == ir.function_name
            and not f.is_shadowed
            and len(f.parameters) == len(ir.arguments)
        ]
        if len(candidates) == 1:
            func = candidates[0]
        if func is None:
            # TODO: handle collision with multiple state variables/functions
            func = contract.get_state_variable_from_name(ir.function_name)
        if func is None and candidates:
            func = _find_function_from_parameter(ir.arguments, candidates, False)

    # low level lookup needs to be done as last step
    if not func:
        if can_be_low_level(ir):
            return convert_to_low_level(ir)
        if can_be_solidity_func(ir):
            return convert_to_solidity_func(ir)
    if not func:
        to_log = f"Function not found {ir.function_name}"
        logger.error(to_log)
    ir.function = func
    if isinstance(func, Function):
        return_type = func.return_type
        # if its not a tuple; return a singleton
        if return_type and len(return_type) == 1:
            return_type = return_type[0]
    else:
        # otherwise its a variable (getter)
        # If its a mapping or a array
        # we iterate until we find the final type
        # mapping and array can be mixed together
        # ex:
        #    mapping ( uint => mapping ( uint => uint)) my_var
        #    mapping(uint => uint)[] test;p
        if isinstance(func.type, (MappingType, ArrayType)):
            tmp = func.type
            while isinstance(tmp, (MappingType, ArrayType)):
                if isinstance(tmp, MappingType):
                    tmp = tmp.type_to
                else:
                    tmp = tmp.type
            return_type = tmp
        else:
            return_type = func.type
    if return_type:
        # If the return type is a structure, but the lvalue is a tuple
        # We convert the type of the structure to a list of element
        # TODO: explore to replace all tuple variables by structures
        if (
            isinstance(ir.lvalue, TupleVariable)
            and isinstance(return_type, UserDefinedType)
            and isinstance(return_type.type, Structure)
        ):
            return_type = _convert_to_structure_to_list(return_type)

        ir.lvalue.set_type(return_type)
    else:
        ir.lvalue = None

    return None


# endregion
###################################################################################
###################################################################################
# region Points to operation
###################################################################################
###################################################################################


def find_references_origin(irs: list[Operation]) -> None:
    """
    Make lvalue of each Index, Member operation
    points to the left variable
    """
    for ir in irs:
        if isinstance(ir, (Index, Member)):
            ir.lvalue.points_to = ir.variable_left


# endregion
###################################################################################
###################################################################################
# region Operation filtering
###################################################################################
###################################################################################


def remove_temporary(result):
    result = [
        ins
        for ins in result
        if not isinstance(
            ins,
            (
                Argument,
                TmpNewElementaryType,
                TmpNewContract,
                TmpNewArray,
                TmpNewStructure,
            ),
        )
    ]

    return result


def remove_unused(result: list[Operation]) -> list[Operation]:
    removed = True

    if not result:
        return result

    # dont remove the last elem, as it may be used by RETURN
    last_elem = result[-1]

    while removed:
        removed = False

        to_keep = []
        to_remove = []

        # keep variables that are read
        # and reference that are written
        for ins in result:
            to_keep += [str(x) for x in ins.read]
            if isinstance(ins, OperationWithLValue) and not isinstance(ins, (Index, Member)):
                if isinstance(ins.lvalue, ReferenceVariable):
                    to_keep += [str(ins.lvalue)]

        for ins in result:
            if isinstance(ins, Member):
                if ins.lvalue.name not in to_keep and ins != last_elem:
                    to_remove.append(ins)
                    removed = True
            # Remove type(X) if X is an elementary type
            # This assume that type(X) is only used with min/max
            # If Solidity introduces other operation, we might remove this removal
            if isinstance(ins, SolidityCall) and ins.function == SolidityFunction("type()"):
                if isinstance(ins.arguments[0], ElementaryType):
                    to_remove.append(ins)

        result = [i for i in result if i not in to_remove]
    return result


# endregion
###################################################################################
###################################################################################
# region Constant type conversion
###################################################################################
###################################################################################


def convert_constant_types(irs: list[Operation]) -> None:
    """
    late conversion of uint -> type for constant (Literal)
    :param irs:
    :return:
    """
    # TODO: implement instances lookup for events, NewContract
    was_changed = True
    while was_changed:
        was_changed = False
        for ir in irs:
            if isinstance(ir, (Assignment, Binary)):
                if (
                    isinstance(ir.lvalue.type, ElementaryType)
                    and ir.lvalue.type.type in ElementaryTypeInt
                ):
                    for r in ir.read:
                        if isinstance(r, Constant) and r.type.type not in ElementaryTypeInt:
                            r.set_type(ElementaryType(ir.lvalue.type.type))
                            was_changed = True

            if isinstance(ir, (HighLevelCall, InternalCall)):
                func = ir.function
                if isinstance(func, StateVariable):
                    types = export_nested_types_from_variable(func)
                else:
                    types = [p.type for p in func.parameters]
                assert len(types) == len(ir.arguments)
                for idx, arg in enumerate(ir.arguments):
                    t = types[idx]
                    if isinstance(t, ElementaryType):
                        if t.type in ElementaryTypeInt:
                            if arg.type.type not in ElementaryTypeInt:
                                arg.set_type(ElementaryType(t.type))
                                was_changed = True

            if isinstance(ir, NewStructure):
                st = ir.structure
                for idx, arg in enumerate(ir.arguments):
                    e = st.elems_ordered[idx]
                    if isinstance(e.type, ElementaryType):
                        if e.type.type in ElementaryTypeInt:
                            if arg.type.type not in ElementaryTypeInt:
                                arg.set_type(ElementaryType(e.type.type))
                                was_changed = True

            def is_elementary_array(t):
                return isinstance(t, ArrayType) and isinstance(t.type, ElementaryType)

            if isinstance(ir, InitArray):
                if is_elementary_array(ir.lvalue.type):
                    if ir.lvalue.type.type.type in ElementaryTypeInt:
                        for r in ir.read:
                            if isinstance(r, Constant) and is_elementary_array(r.type):
                                if r.type.type.type not in ElementaryTypeInt:
                                    r.set_type(ElementaryType(ir.lvalue.type.type.type))
                                    was_changed = True


# endregion
###################################################################################
###################################################################################
# region Delete handling
###################################################################################
###################################################################################


def convert_delete(irs: list[Operation]) -> None:
    """
    Convert the lvalue of the Delete to point to the variable removed
    This can only be done after find_references_origin is called
    :param irs:
    :return:
    """
    for ir in irs:
        if isinstance(ir, Delete):
            if isinstance(ir.lvalue, ReferenceVariable):
                ir.lvalue = ir.lvalue.points_to


# endregion
###################################################################################
###################################################################################
# region Source Mapping
###################################################################################
###################################################################################


def _find_source_mapping_references(irs: list[Operation]) -> None:
    for ir in irs:
        if isinstance(ir, NewContract):
            ir.contract_created.references.append(ir.expression.source_mapping)

        if isinstance(ir, HighLevelCall):
            ir.function.references.append(ir.expression.source_mapping)


# endregion
###################################################################################
###################################################################################
# region Heuristics selection
###################################################################################
###################################################################################


def apply_ir_heuristics(irs: list[Operation], node: "Node", is_solidity: bool) -> list[Operation]:
    """
    Apply a set of heuristic to improve slithIR
    """

    irs = integrate_value_gas(irs)

    irs = propagate_type_and_convert_call(irs, node)
    irs = remove_unused(irs)
    find_references_origin(irs)

    # These are heuristics that are only applied to Solidity
    if is_solidity:
        convert_constant_types(irs)
        convert_delete(irs)

    _find_source_mapping_references(irs)

    return irs



from slither.core.source_mapping.source_mapping import SourceMapping


class Expression(SourceMapping):
    def __init__(self) -> None:
        super().__init__()
        self._is_lvalue = False

    @property
    def is_lvalue(self) -> bool:
        return self._is_lvalue

    def set_lvalue(self) -> None:
        self._is_lvalue = True



from typing import TYPE_CHECKING

from slither.slithir.utils.utils import is_valid_lvalue
from slither.slithir.operations.phi import Phi

from slither.slithir.operations.high_level_call import HighLevelCall
from slither.slithir.operations.internal_call import InternalCall
from slither.slithir.variables.state_variable import StateIRVariable

if TYPE_CHECKING:
    from slither.core.cfg.node import Node


class PhiCallback(Phi):
    def __init__(
        self,
        left_variable: StateIRVariable,
        nodes: set["Node"],
        call_ir: InternalCall | HighLevelCall,
        rvalue: StateIRVariable,
    ) -> None:
        assert is_valid_lvalue(left_variable)
        assert isinstance(nodes, set)
        super().__init__(left_variable, nodes)
        self._call_ir = call_ir
        self._rvalues = [rvalue]
        self._rvalue_no_callback = rvalue

    @property
    def callee_ir(self) -> InternalCall | HighLevelCall:
        return self._call_ir

    @property
    def read(self) -> list[StateIRVariable]:
        return self.rvalues

    @property
    def rvalues(self):
        return self._rvalues

    @rvalues.setter
    def rvalues(self, vals):
        self._rvalues = vals

    @property
    def rvalue_no_callback(self):
        """
        rvalue if callback are not considered
        """
        return self._rvalue_no_callback

    @property
    def nodes(self):
        return self._nodes

    def __str__(self):
        return f"{self.lvalue}({self.lvalue.type}) := \u03d5({[v.ssa_name for v in self._rvalues]})"



import abc
from logging import Logger

from typing import TYPE_CHECKING

from slither.utils import output
from slither.utils.output import SupportedOutput

if TYPE_CHECKING:
    from slither import Slither


class IncorrectPrinterInitialization(Exception):
    pass


class AbstractPrinter(metaclass=abc.ABCMeta):
    ARGUMENT = ""  # run the printer with slither.py --ARGUMENT
    HELP = ""  # help information

    WIKI = ""

    def __init__(self, slither: "Slither", logger: Logger | None) -> None:
        self.slither = slither
        self.contracts = slither.contracts
        self.filename = slither.filename
        self.logger = logger

        if not self.HELP:
            raise IncorrectPrinterInitialization(
                f"HELP is not initialized {self.__class__.__name__}"
            )

        if not self.ARGUMENT:
            raise IncorrectPrinterInitialization(
                f"ARGUMENT is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI:
            raise IncorrectPrinterInitialization(
                f"WIKI is not initialized {self.__class__.__name__}"
            )

    def info(self, info: str) -> None:
        if self.logger:
            self.logger.info(info)

    def generate_output(
        self,
        info: str | list[str | SupportedOutput],
        additional_fields: dict | None = None,
    ) -> output.Output:
        if additional_fields is None:
            additional_fields = {}
        printer_output = output.Output(info, additional_fields)
        printer_output.data["printer"] = self.ARGUMENT

        return printer_output

    @abc.abstractmethod
    def output(self, filename: str) -> output.Output:
        pass



# https://solidity.readthedocs.io/en/v0.4.24/units-and-global-variables.html
from typing import Any

from slither.core.declarations.custom_error import CustomError
from slither.core.solidity_types import ElementaryType, TypeInformation
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.exceptions import SlitherException


SOLIDITY_VARIABLES = {
    "now": "uint256",
    "this": "address",
    "self": "address",
    "abi": "address",  # to simplify the conversion, assume that abi return an address
    "msg": "",
    "tx": "",
    "block": "",
    "super": "",
    "chain": "",
    "ZERO_ADDRESS": "address",
}

SOLIDITY_VARIABLES_COMPOSED = {
    "block.basefee": "uint256",
    "block.blobbasefee": "uint256",
    "block.coinbase": "address",
    "block.difficulty": "uint256",
    "block.prevrandao": "uint256",
    "block.gaslimit": "uint256",
    "block.number": "uint256",
    "block.timestamp": "uint256",
    "block.blockhash": "bytes32",  # alias for blockhash. It's a call
    "block.chainid": "uint256",
    "msg.data": "bytes",
    "msg.gas": "uint256",
    "msg.sender": "address",
    "msg.sig": "bytes4",
    "msg.value": "uint256",
    "tx.gasprice": "uint256",
    "tx.origin": "address",
    # Vyper
    "chain.id": "uint256",
    "block.prevhash": "bytes32",
    "self.balance": "uint256",
}

SOLIDITY_FUNCTIONS: dict[str, list[str]] = {
    "blobhash(uint256)": ["bytes32"],
    "gasleft()": ["uint256"],
    "assert(bool)": [],
    "require(bool)": [],
    "require(bool,string)": [],
    "require(bool,error)": [],  # Solidity 0.8.26 via-ir and Solidity >= 0.8.27
    "revert()": [],
    "revert(string)": [],
    "revert ": [],
    "addmod(uint256,uint256,uint256)": ["uint256"],
    "mulmod(uint256,uint256,uint256)": ["uint256"],
    "keccak256()": ["bytes32"],
    "keccak256(bytes)": ["bytes32"],  # Solidity 0.5
    "sha256()": ["bytes32"],
    "sha256(bytes)": ["bytes32"],  # Solidity 0.5
    "sha3()": ["bytes32"],
    "ripemd160()": ["bytes32"],
    "ripemd160(bytes)": ["bytes32"],  # Solidity 0.5
    "ecrecover(bytes32,uint8,bytes32,bytes32)": ["address"],
    "selfdestruct(address)": [],
    "suicide(address)": [],
    "log0(bytes32)": [],
    "log1(bytes32,bytes32)": [],
    "log2(bytes32,bytes32,bytes32)": [],
    "log3(bytes32,bytes32,bytes32,bytes32)": [],
    "blockhash(uint256)": ["bytes32"],
    "prevrandao()": ["uint256"],
    # the following need a special handling
    # as they are recognized as a SolidityVariableComposed
    # and converted to a SolidityFunction by SlithIR
    "this.balance()": ["uint256"],
    "abi.encode()": ["bytes"],
    "abi.encodePacked()": ["bytes"],
    "abi.encodeWithSelector()": ["bytes"],
    "abi.encodeWithSignature()": ["bytes"],
    "abi.encodeCall()": ["bytes"],
    "bytes.concat()": ["bytes"],
    "string.concat()": ["string"],
    # abi.decode returns a list arbitrary types
    "abi.decode()": [],
    "type(address)": [],
    "type()": [],  # 0.6.8 changed type(address) to type()
    # The following are conversion from address.something
    "balance(address)": ["uint256"],
    "code(address)": ["bytes"],
    "codehash(address)": ["bytes32"],
    # Vyper
    "create_from_blueprint()": [],
    "create_minimal_proxy_to()": [],
    "empty()": [],
    "convert()": [],
    "len()": ["uint256"],
    "method_id()": [],
    "unsafe_sub()": [],
    "unsafe_add()": [],
    "unsafe_div()": [],
    "unsafe_mul()": [],
    "pow_mod256()": [],
    "max_value()": [],
    "min_value()": [],
    "concat()": [],
    "ecrecover()": [],
    "isqrt()": [],
    "range()": [],
    "min()": [],
    "max()": [],
    "shift()": [],
    "abs()": [],
    "raw_call()": ["bool", "bytes32"],
    "_abi_encode()": [],
    "slice()": [],
    "uint2str()": ["string"],
    "send()": [],
}


def solidity_function_signature(name: str) -> str:
    """
        Return the function signature (containing the return value)
        It is useful if a solidity function is used as a pointer
        (see exoressionParsing.find_variable documentation)
    Args:
        name(str):
    Returns:
        str
    """
    return name + f" returns({','.join(SOLIDITY_FUNCTIONS[name])})"


class SolidityVariable(SourceMapping):
    def __init__(self, name: str) -> None:
        super().__init__()
        self._check_name(name)
        self._name = name

    # dev function, will be removed once the code is stable
    def _check_name(self, name: str) -> None:
        assert name in SOLIDITY_VARIABLES or name.endswith(("_slot", "_offset"))

    @property
    def state_variable(self) -> str:
        if self._name.endswith("_slot"):
            return self._name[:-5]
        if self._name.endswith("_offset"):
            return self._name[:-7]
        to_log = f"Incorrect YUL parsing. {self} is not a solidity variable that can be seen as a state variable"
        raise SlitherException(to_log)

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> ElementaryType:
        return ElementaryType(SOLIDITY_VARIABLES[self.name])

    def __str__(self) -> str:
        return self._name

    def __eq__(self, other: Any) -> bool:
        return self.__class__ == other.__class__ and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)


class SolidityVariableComposed(SolidityVariable):
    def _check_name(self, name: str) -> None:
        assert name in SOLIDITY_VARIABLES_COMPOSED

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> ElementaryType:
        return ElementaryType(SOLIDITY_VARIABLES_COMPOSED[self.name])

    def __str__(self) -> str:
        return self._name

    def __eq__(self, other: Any) -> bool:
        return self.__class__ == other.__class__ and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)


class SolidityFunction(SourceMapping):
    # Non standard handling of type(address). This function returns an undefined object
    # The type is dynamic
    # https://solidity.readthedocs.io/en/latest/units-and-global-variables.html#type-information
    # As a result, we set return_type during the Ir conversion

    def __init__(self, name: str) -> None:
        super().__init__()
        assert name in SOLIDITY_FUNCTIONS
        self._name = name
        # Can be TypeInformation if type(address) is used
        self._return_type: list[TypeInformation | ElementaryType] = [
            ElementaryType(x) for x in SOLIDITY_FUNCTIONS[self.name]
        ]

    @property
    def name(self) -> str:
        return self._name

    @property
    def full_name(self) -> str:
        return self.name

    @property
    def return_type(self) -> list[TypeInformation | ElementaryType]:
        return self._return_type

    @return_type.setter
    def return_type(self, r: list[TypeInformation | ElementaryType]) -> None:
        self._return_type = r

    def __str__(self) -> str:
        return self._name

    def __eq__(self, other: Any) -> bool:
        return self.__class__ == other.__class__ and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name)


class SolidityCustomRevert(SolidityFunction):
    def __init__(self, custom_error: CustomError) -> None:
        self._name = "revert " + custom_error.solidity_signature
        self._custom_error = custom_error
        self._return_type: list[TypeInformation | ElementaryType] = []

    @property
    def custom_error(self) -> CustomError:
        return self._custom_error

    def __eq__(self, other: Any) -> bool:
        return (
            self.__class__ == other.__class__
            and self.name == other.name
            and self._custom_error == other._custom_error
        )

    def __hash__(self) -> int:
        return hash(hash(self.name) + hash(self._custom_error))



from slither.core import expressions
from slither.core.expressions.expression import Expression
from slither.visitors.expression.expression import ExpressionVisitor


def get(expression: Expression) -> str:
    val = expression.context["ExpressionPrinter"]
    # we delete the item to reduce memory use
    del expression.context["ExpressionPrinter"]
    return val


def set_val(expression: Expression, val: str) -> None:
    expression.context["ExpressionPrinter"] = val


class ExpressionPrinter(ExpressionVisitor):
    def __init__(self, expression: Expression) -> None:
        self._result: str | None = None
        super().__init__(expression)

    def result(self) -> str:
        if not self._result:
            self._result = get(self.expression)
        return self._result

    def _post_assignement_operation(self, expression: expressions.AssignmentOperation) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        val = f"{left} {expression.type} {right}"
        set_val(expression, val)

    def _post_binary_operation(self, expression: expressions.BinaryOperation) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        val = f"{left} {expression.type} {right}"
        set_val(expression, val)

    def _post_call_expression(self, expression: expressions.CallExpression) -> None:
        called = get(expression.called)
        arguments = ",".join([get(x) for x in expression.arguments if x])
        val = f"{called}({arguments})"
        set_val(expression, val)

    def _post_conditional_expression(self, expression: expressions.ConditionalExpression) -> None:
        if_expr = get(expression.if_expression)
        else_expr = get(expression.else_expression)
        then_expr = get(expression.then_expression)
        val = f"if {if_expr} then {else_expr} else {then_expr}"
        set_val(expression, val)

    def _post_elementary_type_name_expression(
        self, expression: expressions.ElementaryTypeNameExpression
    ) -> None:
        set_val(expression, str(expression.type))

    def _post_identifier(self, expression: expressions.Identifier) -> None:
        set_val(expression, str(expression.value))

    def _post_index_access(self, expression: expressions.IndexAccess) -> None:
        left = get(expression.expression_left)
        right = get(expression.expression_right)
        val = f"{left}[{right}]"
        set_val(expression, val)

    def _post_literal(self, expression: expressions.Literal) -> None:
        set_val(expression, str(expression.value))

    def _post_member_access(self, expression: expressions.MemberAccess) -> None:
        expr = get(expression.expression)
        member_name = str(expression.member_name)
        val = f"{expr}.{member_name}"
        set_val(expression, val)

    def _post_new_array(self, expression: expressions.NewArray) -> None:
        array = str(expression.array_type)
        val = f"new {array}"
        set_val(expression, val)

    def _post_new_contract(self, expression: expressions.NewContract) -> None:
        contract = str(expression.contract_name)
        val = f"new {contract}"
        set_val(expression, val)

    def _post_new_elementary_type(self, expression: expressions.NewElementaryType) -> None:
        t = str(expression.type)
        val = f"new {t}"
        set_val(expression, val)

    def _post_tuple_expression(self, expression: expressions.TupleExpression) -> None:
        underlying_expressions = [get(e) for e in expression.expressions if e]
        val = f"({','.join(underlying_expressions)})"
        set_val(expression, val)

    def _post_type_conversion(self, expression: expressions.TypeConversion) -> None:
        t = str(expression.type)
        expr = get(expression.expression)
        val = f"{t}({expr})"
        set_val(expression, val)

    def _post_unary_operation(self, expression: expressions.UnaryOperation) -> None:
        t = str(expression.type)
        expr = get(expression.expression)
        if expression.is_prefix:
            val = f"{t}{expr}"
        else:
            val = f"{expr}{t}"
        set_val(expression, val)



from crytic_compile import CryticCompile
from slither.core.declarations import (
    Contract,
    FunctionContract,
)
from slither.core.source_mapping.source_mapping import Source, SourceMapping
from slither.exceptions import SlitherError


def get_definition(target: SourceMapping, crytic_compile: CryticCompile) -> Source:
    try:
        pattern = target.pattern
    except AttributeError as exc:
        raise SlitherError(f"get_definition_generic not implemented for {type(target)}") from exc

    file_content = crytic_compile.src_content_for_file(target.source_mapping.filename.absolute)
    txt = file_content[
        target.source_mapping.start : target.source_mapping.start + target.source_mapping.length
    ]

    start_offset = txt.find(pattern) + 1  # remove the space

    starting_line, starting_column = crytic_compile.get_line_from_offset(
        target.source_mapping.filename, target.source_mapping.start + start_offset
    )

    ending_line, ending_column = crytic_compile.get_line_from_offset(
        target.source_mapping.filename, target.source_mapping.start + start_offset + len(pattern)
    )

    s = Source(target.source_mapping.compilation_unit)
    s.start = target.source_mapping.start + start_offset
    s.length = len(pattern)
    s.filename = target.source_mapping.filename
    s.is_dependency = target.source_mapping.is_dependency
    s.lines = list(range(starting_line, ending_line + 1))
    s.starting_column = starting_column
    s.ending_column = ending_column
    s.end = s.start + s.length
    s.txt = txt
    return s


def get_implementation(target: SourceMapping) -> Source:
    return target.source_mapping


def get_all_implementations(target: SourceMapping, contracts: list[Contract]) -> set[Source]:
    """
    Get all implementations of a contract or function, accounting for inheritance and overrides
    """
    implementations = set()
    # Abstract contracts and interfaces are implemented by their children
    if isinstance(target, Contract):
        is_interface = target.is_interface
        is_implicitly_abstract = not target.is_fully_implemented
        is_explicitly_abstract = target.is_abstract
        if is_interface or is_implicitly_abstract or is_explicitly_abstract:
            for contract in contracts:
                if target in contract.immediate_inheritance:
                    implementations.add(contract.source_mapping)

    # Parent's virtual functions may be overridden by children
    elif isinstance(target, FunctionContract):
        for over in target.overridden_by:
            implementations.add(over.source_mapping)
        # Only show implemented virtual functions
        if not target.is_virtual or target.is_implemented:
            implementations.add(get_implementation(target))

    else:
        implementations.add(get_implementation(target))

    return implementations


def get_references(target: SourceMapping) -> list[Source]:
    return target.references



from pathlib import Path
from slither import Slither

TEST_DATA_DIR = Path(__file__).resolve().parent / "test_data"


def test_function_id_rec_structure(solc_binary_path) -> None:
    solc_path = solc_binary_path("0.8.0")
    slither = Slither(Path(TEST_DATA_DIR, "type_helpers.sol").as_posix(), solc=solc_path)
    for compilation_unit in slither.compilation_units:
        for function in compilation_unit.functions:
            assert function.solidity_signature



import sys

try:
    import numpy as np
except ImportError:
    print("ERROR: in order to use slither-simil, you need to install numpy")
    print("$ pip3 install numpy --user\n")
    sys.exit(-1)


def load_cache(infile: str, nsamples: int | None = None) -> dict:
    cache = {}
    with np.load(infile, allow_pickle=True) as data:
        array = data["arr_0"][0]
        for i, (x, y) in enumerate(array):
            cache[x] = y
            if i == nsamples:
                break

    return cache


def save_cache(cache: dict, outfile: str) -> None:
    np.savez(outfile, [np.array(cache)])



from slither.core.declarations import Contract
from slither.core.variables.state_variable import StateVariable
from slither.core.solidity_types import ArrayType, ElementaryType


def is_upgradable_gap_variable(contract: Contract, variable: StateVariable) -> bool:
    """Helper function that returns true if 'variable' is a gap variable used
    for upgradable contracts. More specifically, the function returns true if:
     - variable is named "__gap"
     - it is a uint256 array declared at the end of the contract
     - it has private visibility"""

    # Return early on if the variable name is != gap to avoid iterating over all the state variables
    if variable.name != "__gap":
        return False

    declared_variable_ordered = [
        v for v in contract.state_variables_ordered if v in contract.state_variables_declared
    ]

    if not declared_variable_ordered:
        return False

    variable_type = variable.type
    return (
        declared_variable_ordered[-1] is variable
        and isinstance(variable_type, ArrayType)
        and variable_type.type == ElementaryType("uint256")
        and variable.visibility == "private"
    )



from slither.core.declarations import SolidityFunction, Function
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class ReturnInsteadOfLeave(AbstractDetector):
    """
    Check for cases where a return(a,b) is used in an assembly function that also returns two variables
    """

    ARGUMENT = "return-leave"
    HELP = "If a `return` is used instead of a `leave`."
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#return-instead-of-leave-in-assembly"

    WIKI_TITLE = "Return instead of leave in assembly"
    WIKI_DESCRIPTION = "Detect if a `return` is used where a `leave` should be used."
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract C {
    function f() internal returns (uint a, uint b) {
        assembly {
            return (5, 6)
        }
    }

}
```
The function will halt the execution, instead of returning a two uint."""

    WIKI_RECOMMENDATION = "Use the `leave` statement."

    def _check_function(self, f: Function) -> list[Output]:
        results: list[Output] = []

        for ir in f.solidity_calls:
            if ir.function == SolidityFunction("return(uint256,uint256)"):
                info: DETECTOR_INFO = [f, " contains an incorrect call to return: ", ir.node, "\n"]
                json = self.generate_result(info)

                results.append(json)
        return results

    def _detect(self) -> list[Output]:
        results: list[Output] = []
        for c in self.contracts:
            for f in c.functions_declared:
                if (
                    len(f.returns) == 2
                    and f.contains_assembly
                    and f.visibility not in ["public", "external"]
                ):
                    results += self._check_function(f)

        return results



"""
Module printing summary of the contract
"""

from slither.core.declarations import Function
from slither.core.source_mapping.source_mapping import Source
from slither.printers.abstract_printer import AbstractPrinter
from slither.utils import output
from slither.utils.output import Output


def _get_source_code(cst: Function) -> str:
    src_mapping: Source = cst.source_mapping
    return " " * src_mapping.starting_column + src_mapping.content


class ConstructorPrinter(AbstractPrinter):
    WIKI = "https://github.com/crytic/slither/wiki/Printer-documentation#constructor-calls"
    ARGUMENT = "constructor-calls"
    HELP = "Print the constructors executed"

    def output(self, _filename: str) -> Output:
        info = ""
        for contract in self.slither.contracts_derived:
            stack_name = []
            stack_definition = []
            cst = contract.constructors_declared
            if cst:
                stack_name.append(contract.name)
                stack_definition.append(_get_source_code(cst))
            for inherited_contract in contract.inheritance:
                cst = inherited_contract.constructors_declared
                if cst:
                    stack_name.append(inherited_contract.name)
                    stack_definition.append(_get_source_code(cst))

            if len(stack_name) > 0:
                info += "\n########" + "#" * len(contract.name) + "########\n"
                info += "####### " + contract.name + " #######\n"
                info += "########" + "#" * len(contract.name) + "########\n\n"
                info += "## Constructor Call Sequence" + "\n"

                for name in stack_name[::-1]:
                    info += "\t- " + name + "\n"
                info += "\n## Constructor Definitions" + "\n"
                count = len(stack_definition) - 1
                while count >= 0:
                    info += "\n### " + stack_name[count] + "\n"
                    info += "\n" + str(stack_definition[count]) + "\n"
                    count = count - 1

        self.info(info)
        res = output.Output(info)
        return res
